"""
MIT License

Copyright (c) 2021-present @defxult

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Final,
    Generic,
    Iterable,
    List,
    Literal,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    overload
)

if TYPE_CHECKING:
    from datetime import datetime
    from typing_extensions import Self

import abc
import asyncio
import collections
import inspect
import re
import warnings
from collections.abc import Sequence
from enum import Enum, auto
from typing import NamedTuple, Union

import discord
from discord.ext.commands import Context
from discord.utils import MISSING

from .decorators import ensure_not_primed
from .errors import *

# Constants
_DYNAMIC_EMBED_LIMIT: Final[int] = 4096
_DEFAULT_STYLE: Final[str] = 'Page $/&'
DEFAULT_BUTTONS = None
DEFAULT = MISSING

GB = TypeVar('GB', bound='_BaseButton')
M = TypeVar('M', bound='_BaseMenu')

class Page:
    """Represents a single "page" in the pagination process
    
        .. added:: v3.1.0
    """
    __slots__ = ("content", "embed", "files", "_go_to")
    
    def __init__(self, *, content: Optional[str]=None, embed: Optional[discord.Embed]=MISSING, files: Optional[List[discord.File]]=MISSING) -> None:
        self.content = content
        self.embed = embed
        self.files = files
        self._go_to: Optional[int] = None
    
    def __repr__(self) -> str:
        return f"<Page {' '.join([f'{attr_name}={getattr(self, attr_name)!r}' for attr_name in self.__class__.__slots__ if type(getattr(self, attr_name)) not in (type(None), type(MISSING))])}>"
    
    def _shallow(self) -> Self:
        from copy import copy
        return copy(self)
    
    @staticmethod
    def from_embeds(embeds: Sequence[discord.Embed]) -> List[Page]:
        """|static method|
        
        Converts a sequence of embeds into a list of :class:`Page`
        """
        pages: List[Page] = []
        for e in embeds:
            pages.append(Page(embed=discord.Embed.copy(e)))
        return pages

class PaginationEmojis:
    """A set of basic emojis for your convenience to use for your buttons emoji
    - ◀️ as `BACK_BUTTON`
    - ▶️ as `NEXT_BUTTON`
    - ⏪ as `FIRST_PAGE`
    - ⏩ as `LAST_PAGE`
    - 🔢 as `GO_TO_PAGE`
    - ⏹️ as `END_SESSION`
    """
    BACK_BUTTON: ClassVar[str] = '◀️'
    NEXT_BUTTON: ClassVar[str] = '▶️'
    FIRST_PAGE: ClassVar[str]  = '⏪'
    LAST_PAGE: ClassVar[str]   = '⏩'
    GO_TO_PAGE: ClassVar[str]  = '🔢'
    END_SESSION: ClassVar[str] = '⏹️'

class _PageController:
    def __init__(self, pages: List[Page]) -> None:
        self.ALL_PAGES: Final[List[Page]] = pages
        self.index = 0

    @property
    def current_page(self) -> Page:
        return self.ALL_PAGES[self.index]
    
    @property
    def total_pages(self) -> int:
        """Return the total amount of pages registered to the menu"""
        return len(self.ALL_PAGES) - 1
    
    def validate_index(self) -> Page:
        """If the index is out of bounds, assign the appropriate values so the pagination process can continue and return the associated page"""
        try:
            _ = self.ALL_PAGES[self.index]
        except IndexError:
            if self.index > self.total_pages:
                self.index = 0
            
            elif self.index < 0:
                self.index = self.total_pages
        finally:
            return self.ALL_PAGES[self.index]
    
    def skip_loop(self, action: str, amount: int) -> None:
        """Using `self.index += amount` does not work because this library is used to operating on a +-1 basis. This loop
        provides a simple way to still operate on the +-1 standard.
        """
        while amount != 0:
            if action == '+':
                self.index += 1
            elif action == '-':
                self.index -= 1
            
            self.validate_index()
            amount -= 1
    
    def skip(self, skip: _BaseButton.Skip) -> Page:
        """Return the page that the skip value was set to"""
        self.skip_loop(skip.action, skip.amount)
        return self.validate_index()

    def next(self) -> Page:
        """Return the next page in the pagination process"""
        self.index += 1
        return self.validate_index()
    
    def prev(self) -> Page:
        """Return the previous page in the pagination process"""
        self.index -= 1
        return self.validate_index()
    
    def first_page(self) -> Page:
        """Return the first page in the pagination process"""
        self.index = 0
        return self.ALL_PAGES[self.index]

    def last_page(self) -> Page:
        """Return the last page in the pagination process"""
        self.index = self.total_pages
        return self.ALL_PAGES[self.index]

class _MenuType(Enum):
    TypeEmbed = auto()
    TypeEmbedDynamic = auto()
    TypeText = auto()

MenuType = _MenuType

class _LimitDetails(NamedTuple):
    limit: int
    per: str
    message: str
    set_by_user: bool = False

    @classmethod
    def default(cls) -> Self:
        return cls(0, "", "")

class _BaseButton(Generic[GB], metaclass=abc.ABCMeta):

    Emojis: ClassVar[PaginationEmojis] = PaginationEmojis()

    def __init__(self, name: str, event: Optional[_BaseButton.Event], skip: _BaseButton.Skip):
        self.name: str = name
        self.event: Optional[_BaseButton.Event] = event
        self.skip: _BaseButton.Skip = skip
        self.__clicked_by = set()
        self.__total_clicks = 0
        self.__last_clicked: Optional[datetime] = None
    
    @property
    @abc.abstractmethod
    def menu(self):
        raise NotImplementedError
    
    @property
    def clicked_by(self) -> Set[discord.Member]:
        """
        Returns
        -------
        Set[:class:`discord.Member`]: The members who clicked the button
        """
        return self.__clicked_by
    
    @property
    def total_clicks(self) -> int:
        """
        Returns
        -------
        :class:`int`: The amount of clicks on the button
        """
        return self.__total_clicks

    @property
    def last_clicked(self) -> Optional[datetime]:
        """
        Returns
        -------
        Optional[:class:`datetime.datetime`]: The time in UTC for when the button was last clicked. Can be :class:`None` if the button has not been clicked
        """
        return self.__last_clicked

    def _update_statistics(self, user: Union[discord.Member, discord.User]) -> None:
        self.__clicked_by.add(user)
        self.__total_clicks += 1
        self.__last_clicked = discord.utils.utcnow()

    class Skip:
        """Initialize a skip button with the appropriate values
        
        Parameters
        ----------
        action: :class:`str`
            Whether to go forward in the pagination process ("+") or backwards ("-")
        
        amount: :class:`int`
            The amount of pages to skip. Must be >= 1. If value is <= 0, it is implicitly set to 2
        """
        def __init__(self, action: Literal['+', '-'], amount: int):
            if amount <= 0: amount = 2
            if action in ('+', '-'):
                self.action = action
                self.amount = amount
            else:
                raise MenuException('The action given was not recognized. Expected "+" or "-"')

    class Event:
        """Set a button to be disabled or removed when it has been pressed a certain amount of times. If the button is a :class:`ReactionButton`, only the "remove" event is available
        
        Parameters
        ----------
        event: :class:`str`
            The action to take. Can either be "disable" or "remove"
        
        value: :class:`int`
            The amount set for the specified event. Must be >= 1. If value is <= 0, it is implicitly set to 1"""
        
        _DISABLE = 'disable'
        _REMOVE = 'remove'

        def __init__(self, event_type: Literal['disable', 'remove'], value: int):
            if value <= 0: value = 1
            event_type = str(event_type).lower() # type: ignore
            cls = self.__class__
            
            if event_type in (cls._DISABLE, cls._REMOVE):
                self.event_type = event_type
                self.value = value
            else:
                raise MenuException('The value for parameter "event_type" was not recognized')

class _BaseMenu(metaclass=abc.ABCMeta):
    TypeEmbed: Final[_MenuType] = _MenuType.TypeEmbed
    TypeEmbedDynamic: Final[_MenuType] = _MenuType.TypeEmbedDynamic
    TypeText: Final[_MenuType] = _MenuType.TypeText

    _sessions_limit_details = _LimitDetails.default()
    _active_sessions: List[Self] # initialized in child classes

    def __init__(self, method: Union[Context, discord.Interaction], /, menu_type: _MenuType, **kwargs):
        # `Context` has an `interaction` attribute. If that attribute is `None`,
        # it's not an application command, if so, it's an application command so access
        # and set the interaction attribute so a `ViewMenu` can be used.
        if isinstance(method, Context) and method.interaction is None:
            self._method = method
        elif isinstance(method, Context) and method.interaction is not None:
            self._method = method.interaction
        else:
            self._method = method # will be an interaction

        self._menu_type = menu_type

        self._msg: Union[discord.Message, discord.InteractionMessage] # initialized in child classes
        self._pc:_PageController # initialized in child classes
        self._is_running = False
        self._stop_initiated = False
        self._page_director_separator = ":"

        # dynamic session
        self._main_page_contents = collections.deque()
        self._last_page_contents = collections.deque()
        self._dynamic_data_builder: List[str] = []
        self.wrap_in_codeblock: Optional[str] = kwargs.get('wrap_in_codeblock')
        self.rows_requested: int = kwargs.get('rows_requested', 0)
        self.custom_embed: Optional[discord.Embed] = kwargs.get('custom_embed')

        self._relay_info: Optional[NamedTuple] = None
        self._on_timeout_details: Optional[Callable[[_BaseMenu], None]] = None
        self._menu_timed_out = False
        self._bypass_primed = False # used in :meth:`update()`
        self._pages: List[Page] = []
        self._on_close_event = asyncio.Event() # used for :meth:`wait_until_close()`

        # kwargs
        self.delete_on_timeout: bool = kwargs.get('delete_on_timeout', False)
        self.only_roles: Optional[List[discord.Role]] = kwargs.get('only_roles')
        self.show_page_director: bool = kwargs.get('show_page_director', True)
        self.name: Optional[str] = kwargs.get('name')
        self.style: Optional[str] = kwargs.get('style', _DEFAULT_STYLE)
        self.all_can_click: bool = kwargs.get('all_can_click', False)
        self.delete_interactions: bool = kwargs.get('delete_interactions', True)
        self.allowed_mentions: discord.AllowedMentions = kwargs.get('allowed_mentions', discord.AllowedMentions(everyone=False, users=True, roles=False, replied_user=True))
    
    @abc.abstractmethod
    def _handle_event(self):
        raise NotImplementedError
    
    @abc.abstractmethod
    def remove_all_buttons(self):
        raise NotImplementedError
    
    @abc.abstractmethod
    def get_button(self):
        raise NotImplementedError
    
    @abc.abstractmethod
    def remove_button(self):
        raise NotImplementedError
    
    @abc.abstractmethod
    def add_button(self):
        raise NotImplementedError
    
    @abc.abstractmethod
    def add_buttons(self):
        raise NotImplementedError
    
    @abc.abstractmethod
    def _override_dm_settings(self):
        raise NotImplementedError
    
    @abc.abstractmethod
    def stop(self):
        raise NotImplementedError
    
    @abc.abstractmethod
    async def start(self):
        raise NotImplementedError
    
    # ABC class methods
    
    @abc.abstractmethod
    async def quick_start(cls):
        raise NotImplementedError
    
    @staticmethod
    def separate(values: Sequence[Any]) -> Tuple[List[discord.Embed], List[str]]:
        """|static method|
        
        Sorts all embeds and strings into a single tuple

        Parameters
        ----------
        values: Sequence[`Any`]
            The values to separate
        
        Returns
        -------
        Tuple[List[:class:`discord.Embed`], List[:class:`str`]]
        
        Example
        -------
        >>> embeds, strings = .separate([...])
        """
        all_embeds = list(filter(lambda item: isinstance(item, discord.Embed), values))
        all_strings = list(filter(lambda item: isinstance(item, str), values))
        return (all_embeds, all_strings)
    
    @staticmethod
    def all_embeds(values: Sequence[Any]) -> bool:
        """|static method|
        
        Tests to see if all items in the sequence are of type :class:`discord.Embed`
        
        Parameters
        ----------
        values: Sequence[`Any`]
            The values to test
        
        Returns
        -------
        :class:`bool`: Can return `False` if the sequence is empty
        """
        return all([isinstance(item, discord.Embed) for item in values]) if values else False
    
    @staticmethod
    def _sort_buttons(buttons: List[GB]) -> List[GB]:
        return sorted(buttons, key=lambda btn: btn.total_clicks, reverse=True)
    
    @staticmethod
    def all_strings(values: Sequence[Any]) -> bool:
        """|static method|
        
        Tests to see if all items in the sequence are of type :class:`str`
        
        Parameters
        ----------
        values: Sequence[`Any`]
            The values to test
        
        Returns
        -------
        :class:`bool`: Can return `False` if the sequence is empty
        """
        return all([isinstance(item, str) for item in values]) if values else False
    
    @staticmethod #* Don't make this an instance method. It would be better as one, but it's main intended use is for :meth:`_check` in "views_menu.py"
    def _extract_proper_user(method: Union[Context, discord.Interaction]) -> Union[discord.Member, discord.User]:
        """|static method| Get the proper :class:`discord.User` / :class:`discord.Member` from the attribute depending on the instance

            .. added v3.1.0
        """
        return method.author if isinstance(method, Context) else method.user
    
    @classmethod
    def _quick_check(cls, pages: Sequence[Union[discord.Embed, str]]) -> _MenuType:
        """|class method| Verification for :meth:`quick_start()`
        
            .. added v3.1.0
        """
        if cls.all_embeds(pages):  return cls.TypeEmbed
        if cls.all_strings(pages): return cls.TypeText
        raise IncorrectType(f'All items in the sequence were not of type discord.Embed or str')
    
    @classmethod
    def _all_menu_types(cls) -> Tuple[_MenuType, _MenuType, _MenuType]:
        return (cls.TypeEmbed, cls.TypeEmbedDynamic, cls.TypeText)
    
    @classmethod
    def remove_limit(cls) -> None:
        """|class method|
        
        Remove the limits currently set for menu's
        """
        cls._sessions_limit_details = _LimitDetails.default()
    
    @classmethod
    def get_all_dm_sessions(cls) -> List[Self]:
        """|class method|
        
        Retrieve all active DM menu sessions
        
        Returns
        -------
        A :class:`list` of active DM menu sessions that are currently running. Can be an empty list if there are no active DM sessions
        """
        return [session for session in cls._active_sessions if session._msg.guild is None] # type: ignore
    
    @classmethod
    def get_all_sessions(cls) -> List[Self]:
        """|class method|
        
        Retrieve all active menu sessions
        
        Returns
        -------
        A :class:`list` of menu sessions that are currently running. Can be an empty list if there are no active sessions
        """
        return cls._active_sessions
    
    @classmethod
    def get_session(cls, name: str) -> List[Self]:
        """|class method|
        
        Get a menu instance by it's name
        
        Parameters
        ----------
        name: :class:`str`
            The name of the menu to return
        
        Returns
        -------
        A :class:`list` of menu sessions that are currently running that match the supplied :param:`name`. Can be an empty list if there are no active sessions that matched the :param:`name`
        """
        name = str(name)
        return [session for session in cls._active_sessions if session.name == name]
    
    @classmethod
    def get_sessions_count(cls) -> int:
        """|class method|
        
        Returns the number of active sessions
        
        Returns
        -------
        :class:`int`: The amount of menu sessions that are active
        """
        return len(cls._active_sessions)
    
    @classmethod
    def set_sessions_limit(cls, limit: int, per: Literal['channel', 'guild', 'member']='guild', message: str='Too many active menus. Wait for other menus to be finished.') -> None:
        """|class method|
        
        Sets the amount of menu sessions that can be active at the same time per guild, channel, or member. This applies to both :class:`ReactionMenu` & :class:`ViewMenu`

        Parameters
        ----------
        limit: :class:`int`
            The amount of menu sessions allowed
        
        per: :class:`str`
            How menu sessions should be limited. Options: "channel", "guild", or "member"
        
        message: :class:`str`
            Message that will be sent informing users about the menu limit when the limit is reached. Can be :class:`None` for no message
                    
        Raises
        ------
        - `IncorrectType`: The :param:`limit` parameter was not of type :class:`int`
        - `MenuException`: The value of :param:`per` was not valid or the limit was not greater than or equal to one
        """
        if not isinstance(limit, int):
            raise IncorrectType(f'Parameter "limit" expected int, got {limit.__class__.__name__}')
        else:
            if limit <= 0:
                raise MenuException('The session limit must be greater than or equal to one')
            
            per = str(per).lower() # type: ignore
            if per not in ('guild', 'channel', 'member'):
                raise MenuException('Parameter value of "per" was not recognized. Expected: "channel", "guild", or "member"')

            cls._sessions_limit_details = _LimitDetails(limit, per, message, set_by_user=True)
    
    @classmethod
    async def stop_session(cls, name: str, include_all: bool=False) -> None:
        """|coro class method|
        
        Stop a specific menu with the supplied name
        
        Parameters
        ----------
        name: :class:`str`
            The menus name
        
        include_all: :class:`bool`
            If set to `True`, it stops all menu sessions with the supplied :param:`name`. If `False`, stops only the most recently started menu with the supplied :param:`name`
        
        Raises
        ------
        - `MenuException`: The session with the supplied name was not found
        """
        name = str(name)

        async def determine_include_all(sessions_to_stop: List[Self]) -> None:
            """|coro| Ensures if :param:`include_all` is `True`, all sessions with the specified name will be stopped"""
            if sessions_to_stop:
                if include_all:
                    for session in sessions_to_stop:
                        await session.stop()
                else:
                    await sessions_to_stop[-1].stop()
            else:
                MESSAGE = f'Menu with name {name!r} was not found in the list of active {cls.__name__} sessions'
                raise MenuException(MESSAGE)

        matched_sessions = [session for session in cls._active_sessions if name == session.name]
        await determine_include_all(matched_sessions)
    
    @classmethod
    async def stop_all_sessions(cls) -> None:
        """|coro class method|

        Stops all menu sessions that are currently running
        """
        while cls._active_sessions:
            session = cls._active_sessions[0]
            await session.stop()
    
    @classmethod
    def get_menu_from_message(cls, message_id: int, /) -> Optional[Self]:
        """|class method|

        Return the menu object associated with the message with the given ID

        Parameters
        ----------
        message_id: :class:`int`
            The `discord.Message.id` from the menu message

        Returns
        -------
        The menu object. Can be :class:`None` if the menu was not found in the list of active menu sessions
        """
        for menu in cls._active_sessions:
            if menu._msg.id == message_id: # type: ignore
                return menu
        return None

    @property
    def rows(self) -> Optional[List[str]]:
        """
        Returns
        -------
        Optional[List[:class:`str`]]: All rows that's been added to the menu. Can return `None` if the menu has not started or the `menu_type` is not `TypeEmbedDynamic`
        
            .. added: v3.1.0
        """
        return None if self._is_running is False or self._menu_type != _MenuType.TypeEmbedDynamic else self._dynamic_data_builder.copy()
    
    @property
    def menu_type(self) -> str:
        """
        Returns
        -------
        :class:`str`: The `menu_type` you set via the constructor. This will either be `TypeEmbed`, `TypeEmbedDynamic`, or `TypeText`

            .. added:: v3.1.0
        """
        return self._menu_type.name
    
    @property
    def last_viewed(self) -> Optional[Page]:
        """
        Returns
        -------
        Optional[:class:`Page`]: The last page that was viewed in the pagination process. Can be :class:`None` if the menu has not been started
        """
        return self._pc.current_page._shallow() if self._pc is not None else None
    
    @property
    def owner(self) -> Union[discord.Member, discord.User]:
        """
        Returns
        -------
        Union[:class:`discord.Member`, :class:`discord.User`]: The owner of the menu (the person that started the menu). If the menu was started in a DM, this will return :class:`discord.User`
        """
        return self._extract_proper_user(self._method)
    
    @property
    def total_pages(self) -> int:
        """
        Returns
        -------
        :class:`int`: The amount of pages that have been added to the menu. If the `menu_type` is :attr:`TypeEmbedDynamic`, the amount of pages is not known until AFTER the menu has started. 
        If attempted to retrieve the value before a dynamic menu has started, this will return a value of -1
        """
        if self._menu_type == _BaseMenu.TypeEmbedDynamic:
            return len(self._pages) if self._is_running else -1
        else:
            return len(self._pages)
    
    @property
    def pages(self) -> Optional[List[Page]]:
        """
        Returns
        -------
        Optional[List[:class:`Page`]]: The pages currently applied to the menu. Can return :class:`None` if there are no pages
        
        Note: If the `menu_type` is :attr:`TypeEmbedDynamic`, the pages aren't known until after the menu has started
        """
        return self._pages.copy() if self._pages else None # Return a copy so the core list of pages cannot be manipulated

    @property
    def message(self) -> Optional[Union[discord.Message, discord.InteractionMessage]]:
        """
        Returns
        -------
        Optional[Union[:class:`discord.Message`, :class:`discord.InteractionMessage`]]: The menu's message object. Can be :class:`None` if the menu has not been started
        """
        return self._msg
    
    @property
    def is_running(self) -> bool:
        """
        Returns
        -------
        :class:`bool`: `True` if the menu is currently running, `False` otherwise
        """
        return self._is_running
    
    @property
    def in_dms(self) -> bool:
        """
        Returns
        -------
        :class:`bool`: If the menu was started in a DM
        """
        return self._method.guild is None
    
    def _chunks(self, list_: List[str], n: int) -> Iterable:
        """Yield successive n-sized chunks from list. Core component of a dynamic menu"""
        for i in range(0, len(list_), n):
            yield list_[i:i + n]
    
    async def _build_dynamic_pages(self, send_to, view: Optional[discord.ui.View]=None, payload: Optional[dict]=None) -> None:
        """|coro| Compile all the information that was given via :meth:`add_row`"""
        for data_clump in self._chunks(self._dynamic_data_builder, self.rows_requested):
            joined_data = '\n'.join(data_clump)
            if len(joined_data) <= _DYNAMIC_EMBED_LIMIT:
                possible_block = f"```{self.wrap_in_codeblock}\n{joined_data}```"
                embed = discord.Embed() if self.custom_embed is None else self.custom_embed.copy()
                embed.description = joined_data if not self.wrap_in_codeblock else possible_block
                self._pages.append(Page(embed=embed))
            else:
                raise DescriptionOversized('With the amount of data that was received, the embed description is over discords size limit. Lower the amount of "rows_requested" to solve this problem')
        else:
            def convert_to_page(main_last: Iterable[discord.Embed]) -> List[Page]:
                """Initializing the :class:`deque` only supports :class:`discord.Embed`. This converts those embed objects to the supported :class:`Page` type for proper pagination
                
                    .. added:: v3.1.0
                """
                return [Page(embed=item) for item in main_last]
            
            # set the main/last pages if any
            if any([self._main_page_contents, self._last_page_contents]):
                
                # convert to :class:`deque`
                self._pages = collections.deque(self._pages) # type: ignore (only temporary to use extend methods of :class:`deque`)
                
                if self._main_page_contents:
                    self._main_page_contents.reverse()
                    self._pages.extendleft(convert_to_page(self._main_page_contents)) # type: ignore (converted to deque above)
                
                if self._last_page_contents:
                    self._pages.extend(convert_to_page(self._last_page_contents))
            
            # convert back to `list`
            self._pages = list(self._pages)

            self._refresh_page_director_info(_BaseMenu.TypeEmbedDynamic, self._pages)
            cls = self.__class__

            # make sure data has been added to create at least 1 page
            if not self._pages: raise NoPages(f'You cannot start a {cls.__name__} when no data has been added')
            
            payload['embed'] = self._pages[0].embed # type: ignore / - / reassign the first page to show to director information
            
            if view is None:
                await self._handle_send_to(send_to, payload) # type: ignore
            else:
                await self._handle_send_to(send_to, payload) # type: ignore ("embed=" can still be `None`)
    
    def _display_timeout_warning(self, error: Exception) -> None:
        """Simply displays a warning message to the user notifying them an error has occurred in the function they have set for when the menu times out"""
        warnings.formatwarning = lambda msg, *args, **kwargs: f'{msg}'
        warnings.warn(inspect.cleandoc(
            f"""
            UserWarning: The function you have set in method {self.__class__.__name__}.set_on_timeout() raised an error
            -> {error.__class__.__name__}: {error}
            
            This error has been ignored so the menu timeout process can complete
            """
        ))
    
    async def _handle_on_timeout(self) -> None:
        """|coro| Call the function the user has set for when the menu times out"""
        if self._on_timeout_details and self._menu_timed_out:
            func = self._on_timeout_details
            
            # call the timeout function but ignore any and all exceptions that may occur during the function timeout call.
            # the most important thing here is to ensure the menu is gracefully stopped while displaying a formatted
            # error message to the user
            try:
                if asyncio.iscoroutinefunction(func):
                    await func(self) # type: ignore (`iscoroutinefunction` will still return `False` if func is `None` (i want that to happen))
                else: func(self)
            except Exception as error:
                self._display_timeout_warning(error)
    
    def _determine_kwargs(self, page: Page) -> dict:
        """Determine the `inter.response.edit_message()` and :meth:`_msg.edit()` kwargs for the pagination process. Used in :meth:`ViewMenu._paginate()` and :meth:`ReactionMenu._paginate()`"""
        
        # create a copy of the files if they are set because once paginated, they are only visible once
        maybe_new_files = [discord.File(f.filename) for f in page.files] if page.files else [] # type: ignore
        
        kwargs = {
            "content" : page.content, # `content` will always be present because even with TypeEmbed menu's, pagination with the addition of text is possible
            "allowed_mentions" : self.allowed_mentions,
            "attachments" : maybe_new_files, # the `edit_message` method for this has an "attachment" kwarg instead of a "files" parameter
            "allowed_mentions" : self.allowed_mentions
        }

        # only add the "embed" key if its an embed type menu
        if self._menu_type in (_BaseMenu.TypeEmbed, _BaseMenu.TypeEmbedDynamic):
            kwargs["embed"] = page.embed
        return kwargs
    
    def _refresh_page_director_info(self, type_: _MenuType, pages: List[Page]) -> None:
        """Sets the page count at the bottom of embeds/text if set
        
        Parameters
        ----------
        type_: :class:`_MenuType`
            Either :attr:`_BaseMenu.TypeEmbed`, :attr:`_BaseMenu.TypeEmbedDynamic` or :attr:`_BaseMenu.TypeText`
        
        pages: List[:class:`Page`]
            The pagination contents
        """
        if self.show_page_director:
            if type_ not in (_BaseMenu.TypeEmbed, _BaseMenu.TypeEmbedDynamic, _BaseMenu.TypeText):
                raise Exception('Needs to be of type _BaseMenu.TypeEmbed, _BaseMenu.TypeEmbedDynamic or _BaseMenu.TypeText') 
            
            page_number = 1

            if type_ in (_BaseMenu.TypeEmbed, _BaseMenu.TypeEmbedDynamic):
                page_number = 1
                OUTOF: Final[int] = len(pages)
                all_embeds = [p.embed for p in pages]
                for embed in all_embeds:
                    embed.set_footer(text=f'{self._maybe_new_style(page_number, OUTOF)}{self._page_director_separator if embed.footer.text else ""} {embed.footer.text if embed.footer.text else ""}', icon_url=embed.footer.icon_url) # type: ignore
                    page_number += 1
            else:
                # TypeText Only

                CODEBLOCK = re.compile(r'(`{3})(.*?)(`{3})', flags=re.DOTALL)
                CODEBLOCK_DATA_AFTER = re.compile(r'(`{3})(.*?)(`{3}).+', flags=re.DOTALL)
                for idx in range(len(pages)):
                    page: Page = pages[idx]
                    page_info = self._maybe_new_style(page_number, len(pages))
                    
                    # the main purpose of the re is to decide if only 1 or 2 '\n' should be used. with codeblocks, at the end of the block there is already a new line, so there's no need to add an extra one except in
                    # the case where there is more information after the codeblock
                    
                    # Note: with codeblocks, i already tried the f doc string version of this and it doesnt work because there is a spacing issue with page_info. using a normal f string with \n works as intended
                    # f doc string version: https://github.com/Defxult/reactionmenu/blob/eb88af3a2a6dd468f7bcff38214eb77bc91b241e/reactionmenu/text.py#L288
                    
                    if re.search(CODEBLOCK, page.content): # type: ignore
                        if re.search(CODEBLOCK_DATA_AFTER, page.content): # type: ignore
                            page.content = f'{page.content}\n\n{page_info}'
                        else:
                            page.content = f'{page.content}\n{page_info}'
                    else:
                        page.content = f'{page.content}\n\n{page_info}'
                    page_number += 1
    
    async def _handle_session_limits(self) -> bool:
        """|coro| Determine if the menu session is currently limited, if so, send the error message and return `False` indicating that further code execution (starting the menu) should be cancelled
        
            .. note::
                This method is only called if `_LimitDetails.set_by_user` has been set externally (by the public method)
        """
        cls = self.__class__
        details = cls._sessions_limit_details
        can_proceed = True

        # if the menu is in a DM, handle it separately
        if self.in_dms:
            dm_sessions = cls.get_all_dm_sessions()
            if dm_sessions:
                user_dm_sessions = [session for session in dm_sessions if session.owner.id == self._extract_proper_user(self._method).id]
                if len(user_dm_sessions) >= details.limit:
                    can_proceed = False
        else:
            if details.per == 'guild':
                guild_sessions = [session for session in cls._active_sessions if session.message.guild is not None] # type: ignore
                if len(guild_sessions) >= details.limit:
                    can_proceed = False
            
            elif details.per == 'member':
                member_sessions = [session for session in cls._active_sessions if session.owner.id == self._extract_proper_user(self._method).id]
                if len(member_sessions) >= details.limit:
                    can_proceed = False
            
            elif details.per == 'channel':
                channel_sessions = [session for session in cls._active_sessions if session.message.channel.id == self._method.channel.id] # type: ignore
                if len(channel_sessions) >= details.limit:
                    can_proceed = False
        
        if can_proceed:
            return True
        else:
            await self._method.channel.send(details.message) # type: ignore
            return False
    
    def _maybe_new_style(self, counter: int, total_pages: int) -> str: 
        """Sets custom page director styles"""
        if self.style:
            if self.style.count('$') == 1 and self.style.count('&') == 1:
                temp = self.style # copy it to a new variable so its not being changed in every call
                temp = temp.replace('$', str(counter))
                temp = temp.replace('&', str(total_pages))
                return temp
            else:
                raise ImproperStyleFormat
        else:
            return f'Page {counter}/{total_pages}'
    
    async def _contact_relay(self, member: Union[discord.Member, discord.User], button: GB) -> None: # type: ignore
        """|coro| Dispatch the information to the relay function if a relay has been set"""
        if self._relay_info:
            func: Callable = self._relay_info.func # type: ignore
            only = self._relay_info.only # type: ignore
            RelayPayload = collections.namedtuple('RelayPayload', ['member', 'button'])
            payload = RelayPayload(member=member, button=button)

            # before the information is relayed, ensure the relay function contains a single positional argument
            spec = inspect.getfullargspec(func)
            if all([
                len(spec.args) == 1,
                not spec.varargs,
                not spec.varkw,
                not spec.defaults,
                not spec.kwonlyargs,
                not spec.kwonlydefaults
            ]):
                async def call() -> None:
                    """|coro| Dispatch the information to the relay function. If any errors occur during the call, report it to the user"""
                    try:
                        if asyncio.iscoroutinefunction(func):
                            await func(payload)
                        else:
                            func(payload)
                    except Exception as error:
                        error_msg = inspect.cleandoc(
                            f"""When dispatching the information to your relay function ("{func.__name__}"), that function raised an error during it's execution
                            -> {error.__class__.__name__}: {error}
                            """
                        )
                        raise MenuException(error_msg)

                if only:
                    if button in only:
                        await call()
                else:
                    await call()
            else:
                raise MenuException('When setting a relay, the relay function must have exactly one positional argument')
    
    def _handle_reply_kwargs(self, send_to, reply: bool) -> dict:
        """Used to determine the mentions for the `reply` parameter in :meth:`.start()`"""
        if isinstance(send_to, (discord.TextChannel, discord.VoiceChannel, discord.Thread)) and send_to == self._method.channel:
            send_to = None
        return {
            'reference' : self._method.message if all([send_to is None, reply is True]) else None,
            'mention_author' : self.allowed_mentions.replied_user
        }
    
    async def _handle_send_to(self, send_to: Union[str, int, discord.TextChannel, discord.VoiceChannel, discord.Thread, None], menu_payload: dict) -> None:
        """|coro| For the :param:`send_to` kwarg in :meth:`Menu.start()`. Determine what channel the menu should start in"""
        if isinstance(self._method, Context):

            async def register_message(channel: discord.abc.Messageable) -> None:
                """|coro| Set :attr:`_msg` to the :class:`discord.Message` that the menu is operating from"""
                self._msg = await channel.send(**menu_payload) # type: ignore
            
            if self.in_dms:
                await register_message(self._method.channel)
            else:
                if send_to is None:
                    await register_message(self._method.channel)
                else:
                    if not isinstance(send_to, (str, int, discord.TextChannel, discord.VoiceChannel, discord.Thread)):
                        raise IncorrectType(f'Parameter "send_to" expected str, int, discord.TextChannel, discord.VoiceChannel, or discord.Thread, got {send_to.__class__.__name__}')
                    else:
                        class_name = self.__class__.__name__                    # converted to list because its a sequence proxy
                        all_messageable_channels = self._method.guild.text_channels + list(self._method.guild.threads) + self._method.guild.voice_channels # type: ignore
                        
                        # before we continue, check if there are any duplicate named text channels or threads/no matching names found if a str was provided
                        if isinstance(send_to, str):
                            matched_channels = [ch for ch in all_messageable_channels if ch.name == send_to]
                            if len(matched_channels) == 0:
                                raise MenuException(f'When using parameter "send_to" in {class_name}.start(), there were no channels/threads with the name {send_to!r}')
                            
                            elif len(matched_channels) >= 2:
                                raise MenuException(f'When using parameter "send_to" in {class_name}.start(), there were {len(matched_channels)} channels/threads with the name {send_to!r}. With multiple channels/threads having the same name, the intended channel is unknown')
                        
                        for channel in all_messageable_channels:
                            if isinstance(send_to, str):
                                if channel.name == send_to:
                                    await register_message(channel)
                                    break
                            
                            elif isinstance(send_to, int):
                                if channel.id == send_to:
                                    await register_message(channel)
                                    break
                            
                            # it should be a discord.TextChannel, discord.VoiceChannel, or discord.Thread
                            else:
                                if channel == send_to:
                                    await register_message(channel)
                                    break

                        else:
                            raise MenuException(f'When using parameter "send_to" in {class_name}.start(), the channel {send_to} was not found')
        
        elif isinstance(self._method, discord.Interaction):
            if self._method.response.is_done():
                await self._method.followup.send(**menu_payload)
            else:
                await self._method.response.send_message(**menu_payload)
            self._msg = await self._method.original_response()
        
        else:
            raise IncorrectType('Parameter "method" was not of the correct type')

    def randomize_embed_colors(self) -> None:
        """Randomize the color of all the embeds that have been added to the menu
        
        Raises
        ------
        - `MenuException`: The `menu_type` was not of `TypeEmbed`
        
            .. added:: v3.1.0
        """
        if self._menu_type == _BaseMenu.TypeEmbed:
            for page in self._pages:
                if page.embed:
                    page.embed.color = discord.Color.random()
        else:
            raise MenuException("Randomizing embed colors can only be used when the menu_type is TypeEmbed")
    
    def set_page_director_style(self, style_id: int, separator: str=DEFAULT) -> None:
        """Set how the page numbers dictating what page you are on (in the footer of an embed/regular message) are displayed

        Parameters
        ----------
        style_id: :class:`int`
            Varying formats of how the page director can be presented. The following ID's are available:

            - `1` = Page 1/10
            - `2` = Page 1 out of 10
            - `3` = 1 out of 10
            - `4` = 1 • 10
            - `5` = 1 » 10
            - `6` = 1 | 10
            - `7` = 1 : 10
            - `8` = 1 - 10
            - `9` = 1 / 10
            - `10` = 1 🔹 10
            - `11` = 1 🔸 10
        
        separator: :class:`str`
            The separator between the page director and any text you may have in the embed footer. The default separator is ":". It should be noted that whichever separator you assign,
            if you wish to have spacing between the page director and the separator, you must place the space inside the string yourself as such: " :"
        
        Raises
        ------
        - `MenuException`: The :param:`style_id` value was not valid 
        """
        if separator:
            self._page_director_separator = str(separator)
        
        if style_id == 1:   self.style = _DEFAULT_STYLE
        elif style_id == 2: self.style = 'Page $ out of &'
        elif style_id == 3: self.style = '$ out of &'
        elif style_id == 4: self.style = '$ • &'
        elif style_id == 5: self.style = '$ » &'
        elif style_id == 6: self.style = '$ | &'
        elif style_id == 7: self.style = '$ : &'
        elif style_id == 8: self.style = '$ - &'
        elif style_id == 9: self.style = '$ / &'
        elif style_id == 10: self.style = '$ 🔹 &'
        elif style_id == 11: self.style = '$ 🔸 &'
        else:
            raise MenuException(f'Parameter "style_id" expected a number 1-11, got {style_id!r}')
    
    async def wait_until_closed(self) -> None:
        """|coro|
        
        Waits until the menu session ends using `.stop()` or when the menu times out. This should not be used inside relays
        
            .. added:: v3.0.1
        """
        await self._on_close_event.wait()
    
    @ensure_not_primed
    def add_from_messages(self, messages: Sequence[discord.Message]) -> None:
        """Add pages to the menu using the message object itself
        
        Parameters
        ----------
        messages: Sequence[:class:`discord.Message`]
            A sequence of discord message objects
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `MenuSettingsMismatch`: The messages provided did not have the correct values. For example, the `menu_type` was set to `TypeEmbed`, but the messages you've provided only contains text. If the `menu_type` is `TypeEmbed`, only messages with embeds should be provided
        - `IncorrectType`: All messages were not of type :class:`discord.Message`
        """
        if all([isinstance(msg, discord.Message) for msg in messages]):
            if self._menu_type == _BaseMenu.TypeEmbed:
                embeds: List[discord.Embed] = []
                for m in messages:
                    if m.embeds:
                        embeds.extend(m.embeds)
                if embeds:
                    self.add_pages(embeds)
                else:
                    raise MenuSettingsMismatch(f'The menu is set to {self._menu_type.name} but no embeds were found in the messages provided')
            
            elif self._menu_type == _BaseMenu.TypeText:
                content: List[str] = []
                for m in messages:
                    if m.content:
                        content.append(m.content)
                if content:
                    self.add_pages(content)
                else:
                    raise MenuSettingsMismatch(f'The menu is set to {self._menu_type.name} but no text (discord.Message.content) was found in the messages provided')
        else:
            raise IncorrectType('All messages were not of type discord.Message')
    
    @ensure_not_primed
    async def add_from_ids(self, messageable: discord.abc.Messageable, message_ids: Sequence[int]) -> None:
        """|coro|
        
        Add pages to the menu using the IDs of messages. This only grabs embeds (if the `menu_type` is :attr:`TypeEmbed`) or the content (if the `menu_type` is :attr:`TypeText`) from the message
        
        Parameters
        ----------
        messageable: :class:`discord.abc.Messageable`
            A discord :class:`Messageable` object (:class:`discord.TextChannel`, :class:`commands.Context`, etc.)
        
        message_ids: Sequence[:class:`int`]
            The messages to fetch

        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `MenuSettingsMismatch`: The message IDs provided did not have the correct values when fetched. For example, the `menu_type` was set to `TypeEmbed`, but the messages you've provided for the library to fetch only contains text. If the `menu_type` is `TypeEmbed`, only messages with embeds should be provided
        - `MenuException`: An error occurred when attempting to fetch a message or not all :param:`message_ids` were of type int
        """
        if all([isinstance(ID, int) for ID in message_ids]):
            to_paginate: List[discord.Message] = []            
            for msg_id in message_ids:
                try:
                    fetched_message = await messageable.fetch_message(msg_id)
                    to_paginate.append(fetched_message)
                except (discord.NotFound, discord.Forbidden, discord.HTTPException) as error:
                    raise MenuException(f'An error occurred when attempting to retrieve message with the ID {msg_id}: {error}')
            
            if self._menu_type == _BaseMenu.TypeEmbed:
                embeds_to_paginate: List[discord.Embed] = []
                for msg in to_paginate:
                    if msg.embeds:
                        embeds_to_paginate.extend(msg.embeds)
                if embeds_to_paginate:
                    for embed in embeds_to_paginate:
                        self.add_page(embed)
                else:
                    raise MenuSettingsMismatch(f'The menu is set to {self._menu_type.name} but no embeds were found in the messages provided')
            
            elif self._menu_type == _BaseMenu.TypeText:
                content_to_paginate: List[str] = []
                for msg in to_paginate:
                    if msg.content:
                        content_to_paginate.append(msg.content)
                if content_to_paginate:
                    for content in content_to_paginate:
                        self.add_page(content=content)
                else:
                    raise MenuSettingsMismatch(f'The menu is set to {self._menu_type.name} but no text (discord.Message.content) was found in the messages provided')
        else:
            raise MenuException('Not all message IDs were of type int')
    
    @ensure_not_primed
    def clear_all_row_data(self) -> None:
        """Delete all the data thats been added using :meth:`add_row()`
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `MenuSettingsMismatch`: This method was called but the menus `menu_type` was not :attr:`TypeEmbedDynamic`
        """
        if self._menu_type == _BaseMenu.TypeEmbedDynamic:
            self._dynamic_data_builder.clear()
        else:
            raise MenuSettingsMismatch('Cannot clear all row data when the menu_type is not set as TypeEmbedDynamic')
    
    @ensure_not_primed
    def add_row(self, data: str) -> None:
        """Add text to the embed page by rows of data
        
        Parameters
        ----------
        data: :class:`str`
            The data to add
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `MenuSettingsMismatch`: This method was called but the menus `menu_type` was not :attr:`TypeEmbedDynamic`
        - `MissingSetting`: The kwarg "rows_requested" (int) has not been set for the menu
        """
        if self._menu_type == _BaseMenu.TypeEmbedDynamic:
            if self.rows_requested:
                self._dynamic_data_builder.append(str(data))
            else:
                raise MissingSetting('The kwarg "rows_requested" (int) has not been set for the menu')
        else:
            raise MenuSettingsMismatch('add_row() can only be used with a menu_type of TypeEmbedDynamic')
    
    @ensure_not_primed
    def set_main_pages(self, *embeds: discord.Embed) -> None:
        """On a menu with a `menu_type` of :attr:`TypeEmbedDynamic`, set the pages you would like to show first. These embeds will be shown before the embeds that contain your data
        
        Parameters
        ----------
        *embeds: :class:`discord.Embed`
            An argument list of :class:`discord.Embed` objects
        
        Raises
        ------
        - `MenuSettingsMismatch`: Tried to use method on a menu that was not of `menu_type` :attr:`TypeEmbedDynamic`
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `MenuException`: The "embeds" parameter was empty. At least one value is needed
        - `IncorrectType`: All values in the argument list were not of type :class:`discord.Embed`
        """
        if not embeds: raise MenuException('The argument list when setting main pages was empty')
        if not all([isinstance(e, discord.Embed) for e in embeds]): raise IncorrectType('All values in the argument list when setting main pages were not of type discord.Embed')
        if self._menu_type != _BaseMenu.TypeEmbedDynamic: raise MenuSettingsMismatch('Method set_main_pages is only available for menus with menu_type TypeEmbedDynamic')
        
        # if they've set any values, remove it. Each set should be from the call and should not stack
        self._main_page_contents.clear()
        
        for embed in embeds:
            self._main_page_contents.append(embed)

    @ensure_not_primed
    def set_last_pages(self, *embeds: discord.Embed) -> None:
        """On a menu with a `menu_type` of :attr:`TypeEmbedDynamic`, set the pages you would like to show last. These embeds will be shown after the embeds that contain your data
        
        Parameters
        ----------
        *embeds: :class:`discord.Embed`
            An argument list of :class:`discord.Embed` objects
        
        Raises
        ------
        - `MenuSettingsMismatch`: Tried to use method on a menu that was not of `menu_type` :attr:`TypeEmbedDynamic`
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `MenuException`: The "embeds" parameter was empty. At least one value is needed
        - `IncorrectType`: All values in the argument list were not of type :class:`discord.Embed`
        """
        if not embeds: raise MenuException('The argument list when setting last pages was empty')
        if not all([isinstance(e, discord.Embed) for e in embeds]): raise IncorrectType('All values in the argument list when setting last pages were not of type discord.Embed')
        if self._menu_type != _BaseMenu.TypeEmbedDynamic: raise MenuSettingsMismatch('Method set_last_pages is only available for menus with menu_type TypeEmbedDynamic')
        
        # if they've set any values, remove it. Each set should be from the call and should not stack
        self._last_page_contents.clear()
        
        for embed in embeds:
            self._last_page_contents.append(embed)
    
    @ensure_not_primed
    def add_page(self, embed: Optional[discord.Embed]=MISSING, content: Optional[str]=None, files: Optional[List[discord.File]]=MISSING) -> None:
        """Add a page to the menu

        Parameters
        ----------
        embed: Optional[:class:`discord.Embed`]
            The embed of the page
        
        content: Optional[:class:`str`]
            The text that appears above an embed in a message
        
        files: Optional[Sequence[:class:`discord.File`]]
            Files you'd like to attach to the page
        
        Raises
        ------
        - `MenuException`: Attempted to add a page with no parameters
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `MenuSettingsMismatch`: The page being added does not match the menus `menu_type` 
        
            .. changes::
                v3.1.0
                    Added parameter content
                    Added parameter embed
                    Added parameter files
                    Removed parameter "page"
        """
        if all([content is None, embed is None, files is None]):
            raise MenuException("When adding a page, at lease one parameter must be set")
        
        cls = self.__class__
        if self._menu_type == cls.TypeEmbed:
            if isinstance(embed, discord.Embed):
                self._pages.append(Page(content=content, embed=embed, files=files))
            else:
                raise MenuSettingsMismatch(f'When adding a page with a menu_type of TypeEmbed, the "embed" parameter cannot be None')
        
        elif self._menu_type == cls.TypeText:
            if content:
                self._pages.append(Page(content=str(content), files=files))
            else:
                raise MenuSettingsMismatch(f'When adding a page with a menu_type of TypeText, the "content" parameter cannot be None')
        
        else:
            raise MenuSettingsMismatch('add_page method cannot be used with the current menu_type')
    
    @overload
    def add_pages(self, pages: Sequence[discord.Embed]) -> None:
        ...
    
    @overload
    def add_pages(self, pages: Sequence[str]) -> None:
        ...

    @ensure_not_primed
    def add_pages(self, pages: Sequence[Union[discord.Embed, str]]) -> None:
        """Add multiple pages to the menu at once
        
        Parameters
        ----------
        pages: Sequence[Union[:class:`discord.Embed`, :class:`str`]]
            The pages to add. Can only be used when the menus `menu_type` is :attr:`TypeEmbed` (adding a :class:`discord.Embed`)
            or :attr:`TypeText` (adding a :class:`str`)
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `MenuSettingsMismatch`: The page being added does not match the menus `menu_type` 
        """
        for embed_or_str in pages:
            if isinstance(embed_or_str, str):
                self.add_page(content=embed_or_str)
            else:
                self.add_page(embed_or_str) # type: ignore
    
    @ensure_not_primed
    def remove_all_pages(self) -> None:
        """Remove all pages from the menu
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        """
        self._pages.clear()
    
    @ensure_not_primed
    def remove_page(self, page_number: int) -> None:
        """Remove a page from the menu
        
        Parameters
        ----------
        page_number: :class:`int`
            The page to remove
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `InvalidPage`: The page associated with the given page number was not valid
        """
        if self._pages:
            if page_number > 0 and page_number <= len(self._pages):
                page_to_delete = page_number - 1
                del self._pages[page_to_delete]
            else:
                raise InvalidPage(f'Page number invalid. Must be from 1 - {len(self._pages)}')
    
    def set_on_timeout(self, func: Callable[[M], None]) -> None:
        """Set the function to be called when the menu times out

        Parameters
        ----------
        func: Callable[[:type:`M`]], :class:`None`]
            The function object that will be called when the menu times out. The function should contain a single positional argument
            and should not return anything. The argument passed to that function is an instance of the menu.
        
        Raises
        ------
        - `IncorrectType`: Parameter "func" was not a callable object
        """
        if not callable(func): raise IncorrectType('Parameter "func" must be callable')
        self._on_timeout_details = func # type: ignore
    
    def remove_on_timeout(self) -> None:
        """Remove the timeout call to the function you have set when the menu times out"""
        self._on_timeout_details = None
    
    def set_relay(self, func: Callable[[NamedTuple], None], *, only: Optional[List[GB]]=None) -> None:
        """Set a function to be called with a given set of information when a button is pressed on the menu. The information passed is `RelayPayload`, a named tuple.
        The named tuple contains the following attributes:

        - `member`: The :class:`discord.Member` object of the person who pressed the button. Could be :class:`discord.User` if the menu was started in a DM
        - `button`: Depending on the menu instance, the :class:`ReactionButton` or :class:`ViewButton` object of the button that was pressed

        Parameters
        ----------
        func: Callable[[:class:`NamedTuple`], :class:`None`]
            The function should only contain a single positional argument. Command functions (`@bot.command()`) not supported
        
        only: Optional[List[:generic:`GB`]]
            A list of buttons (`GB`) associated with the current menu instance. If the menu instance is :class:`ReactionMenu`, this should be a list of :class:`ReactionButton`
            and vice-versa for :class:`ViewMenu` instances. If this is :class:`None`, all buttons on the menu will be relayed. If set, only button presses from those specified buttons will be relayed
        
        Raises
        ------
        - `IncorrectType`: The :param:`func` argument provided was not callable
        """
        if callable(func):
            RelayInfo = collections.namedtuple('RelayInfo', ['func', 'only'])
            self._relay_info = RelayInfo(func=func, only=only)
        else:
            raise IncorrectType('When setting the relay, argument "func" must be callable')
    
    def remove_relay(self) -> None:
        """Remove the relay that's been set"""
        self._relay_info = None



"""
reactionmenu • discord pagination
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A library to create a discord.py 2.0+ paginator. Supports pagination with buttons, reactions, and category selection using selects.

:copyright: (c) 2021-present @defxult
:license: MIT

"""

from .buttons import ReactionButton, ViewButton
from .core import ReactionMenu
from .views_menu import ViewMenu, ViewSelect
from .abc import Page


__source__ = 'https://github.com/Defxult/reactionmenu'
__all__ = (
    'ReactionMenu',
    'ReactionButton',
    'ViewMenu',
    'ViewButton',
    'ViewSelect',
    'Page'
)



"""
MIT License

Copyright (c) 2021-present @defxult

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import asyncio
import collections
import inspect
import random
import re
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Final,
    List,
    Literal,
    NamedTuple,
    NoReturn,
    Optional,
    Sequence,
    Union,
    overload
)

if TYPE_CHECKING:
    from .abc import MenuType

import discord
from discord.ext.commands import Context

from . import ViewButton
from .abc import (
    _DEFAULT_STYLE,
    DEFAULT_BUTTONS,
    Page,
    _BaseMenu,
    _PageController
)

from .decorators import ensure_not_primed
from .errors import *

_SelectOptionRelayPayload = collections.namedtuple('_SelectOptionRelayPayload', ['func', 'only'])

class ViewSelect(discord.ui.Select):
    """A class to assist in the process of categorizing information on a :class:`ViewMenu`

    Parameters
    ----------
    title: Union[:class:`str`, `None`]
        The category name. If `None`, the category name defaults to "Select a category"
    
    options: Dict[:class:`discord.SelectOption`, List[:class:`Page`]]
        A key/value pair associating the category options with pages to navigate
    
    disabled: :class:`bool`
        If the select should start enabled or disabled

        .. added:: v3.1.0
    """
    def __init__(self, *, title: Union[str, None], options: Dict[discord.SelectOption, List[Page]], disabled: bool=False) -> None:
        self._menu: Optional[ViewMenu] = None
        self._view_select_options = options
        
        super().__init__(custom_id=discord.utils.MISSING, placeholder='Select a category' if title is None else title, min_values=1, max_values=1, options=options.keys(), disabled=disabled, row=None) # type: ignore
    
    def __repr__(self) -> str:
        return f'<ViewSelect title={self.placeholder!r} disabled={self.disabled}>'
    
    @property
    def menu(self) -> Optional[ViewMenu]:
        """
        Returns
        -------
        Optional[:class:`ViewMenu`]: The menu this select is attached to. Can be `None` if not attached to any menu
        """
        return self._menu
    
    async def __dispatch_relay(self, interaction: discord.Interaction) -> None:
        if self._menu:
            relay_info = self._menu._options_relay_info
            if relay_info is not None:
                SelectOptionRelayPayload = collections.namedtuple('SelectOptionRelayPayload', ['member', 'option', 'menu'])
                CLICKED_OPTION: Final[str] = self.values[0]
                
                # Find which :class:`discord.SelectOption` was clicked
                filtered_options = [sel for sel in self._view_select_options if sel.label == CLICKED_OPTION]
                select_option = filtered_options[0]
                payload = SelectOptionRelayPayload(member=interaction.user, option=select_option, menu=self._menu)

                async def call():
                    if asyncio.iscoroutinefunction(relay_info.func):
                        await relay_info.func(payload)
                    else:
                        relay_info.func(payload)
                
                if relay_info.only:
                    for option_label in relay_info.only:
                        if CLICKED_OPTION == option_label:
                            await call()
                            break
                else:
                    if filtered_options:
                        await call()
    
    async def callback(self, interaction: discord.Interaction) -> None:
        """*INTERNAL USE ONLY* - The callback function from the select interaction. This should not be manually called"""
        for option, pages in self._view_select_options.items():
            if option.label == self.values[0]:
                if self._menu:
                    self._menu._pc = _PageController(pages)
                    first_page = self._menu._pc.first_page()
                    await interaction.response.edit_message(**self._menu._determine_kwargs(first_page))
                    break
        await self.__dispatch_relay(interaction)

    class GoTo(discord.ui.Select):
        """Represents a UI based version of a :class:`ViewButton` with the ID `ViewButton.ID_GO_TO_PAGE`

        Parameters
        ----------
        title: Union[:class:`str`, `None`]
            The selects name. If `None`, the name defaults to "Navigate to page..."
        
        page_numbers: This parameter accepts 3 different types and are explained below
        - List[:class:`int`]
            - If set to a list of integers, those specified values are the only options that are available when the select is clicked
        
        - Dict[:class:`int`, Union[:class:`str`, :class:`discord.Emoji`, :class:`discord.PartialEmoji`]]
            - You can use this type if you'd like to utilize emojis in your select options
        
        - `ellipsis`
            - Set an ellipsis to have the library automatically assign all page numbers to the amount of pages that you've added to the menu.
            
        NOTE: Setting the `page_numbers` parameter to an ellipsis (...) only works as intended if you've added the go to select AFTER you've added pages to the menu
        
            .. added:: v3.1.0
        """
        def __init__(self, *, title: Union[str, None], page_numbers: Union[List[int], Dict[int, Union[str, discord.Emoji, discord.PartialEmoji]], ellipsis]) -> None:
            self.callback = self._select_go_to_callback
            self._menu: Optional[ViewMenu] = None
            self._page_numbers = page_numbers
            self._gt_options: List[discord.SelectOption] = []
            
            if isinstance(page_numbers, list):
                self._gt_options.extend([discord.SelectOption(label=str(n)) for n in page_numbers])
            elif isinstance(page_numbers, dict):
                for n, emoji in page_numbers.items():
                    self._gt_options.append(discord.SelectOption(label=str(n), emoji=emoji))
            
            super().__init__(placeholder="Navigate to page..." if title is None else str(title), options=self._gt_options)
        
        @property
        def menu(self) -> Optional[ViewMenu]:
            """
            Returns
            -------
            Optional[:class:`ViewMenu`]: The menu this select is attached to. Can be `None` if not attached to any menu
            """
            return self._menu
        
        async def _select_go_to_callback(self, interaction: discord.Interaction) -> None:
            if self._menu._check(interaction): # type: ignore
                if interaction.data:
                    values = interaction.data.get('values') 
                    if values:
                        CLICKED_OPTION: Final[int] = int(values[0])
                        if self._menu:
                            if 1 <= CLICKED_OPTION <= self._menu._pc.total_pages + 1:
                                self._menu._pc.index = CLICKED_OPTION - 1
                                await interaction.response.edit_message(**self._menu._determine_kwargs(self._menu._pc.current_page))
                                return
                else:
                    assert False, "No interaction data"
            else:
                await interaction.response.defer()

class ViewMenu(_BaseMenu):
    """A class to create a discord pagination menu using :class:`discord.ui.View`
    
    Parameters
    ----------
    method: Union[:class:`discord.ext.commands.Context`, :class:`discord.Interaction`]
        The Context object. You can get this using a command or if you're in a `discord.on_message` event. Also accepts interactions, typically received when using slash commands
    
    menu_type: :class:`MenuType`
        The configuration of the menu. Class variables :attr:`ViewMenu.TypeEmbed`, :attr:`ViewMenu.TypeEmbedDynamic`, or :attr:`ViewMenu.TypeText`
    
    Kwargs
    ------
    all_can_click: :class:`bool`
        Sets if everyone is allowed to control when pages are 'turned' when buttons are pressed (defaults to `False`)
    
    allowed_mentions: :class:`discord.AllowedMentions`
        Controls the mentions being processed in the menu message (defaults to :class:`discord.AllowedMentions(everyone=False, users=True, roles=False, replied_user=True)`).
        Not valid for `ViewMenu` with a `menu_type` of `TypeText`
    
    custom_embed: :class:`discord.Embed`
        Embed object to use when adding data with :meth:`ViewMenu.add_row()`. Used for styling purposes only (:attr:`ViewMenu.TypeEmbedDynamic` only/defaults to :class:`None`)
    
    delete_interactions: :class:`bool`
        Delete the prompt message by the bot and response message by the user when asked what page they would like to go to when using :attr:`ViewButton.ID_GO_TO_PAGE` (defaults to `True`)
    
    delete_on_timeout: :class:`bool`
        Delete the menu when it times out (defaults to `False`) If `True`, :attr:`disable_items_on_timeout` and :attr:`remove_items_on_timeout` will not execute regardless of if they are `True`. This takes priority over those actions
    
    disable_items_on_timeout: :class:`bool`
        Disable the buttons on the menu when the menu times out (defaults to `True`) If :attr:`delete_on_timeout` is `True`, this will be overridden
    
    name: :class:`str`
        A name you can set for the menu (defaults to :class:`None`)
    
    only_roles: List[:class:`discord.Role`]
        If set, only members with any of the given roles are allowed to control the menu. The menu owner can always control the menu (defaults to :class:`None`)
    
    remove_items_on_timeout: :class:`bool`
        Remove the buttons on the menu when the menu times out (defaults to `False`) If :attr:`disable_items_on_timeout` is `True`, this will be overridden
    
    rows_requested: :class:`int`
        The amount of information per :meth:`ViewMenu.add_row()` you would like applied to each embed page (:attr:`ViewMenu.TypeEmbedDynamic` only/defaults to :class:`None`)
    
    show_page_director: :class:`bool`
        Shown at the bottom of each embed page. "Page 1/20" (defaults to `True`)
    
    style: :class:`str`
        A custom page director style you can select. "$" represents the current page, "&" represents the total amount of pages (defaults to "Page $/&") Example: `ViewMenu(ctx, ..., style='On $ out of &')`
    
    timeout: Union[:class:`int`, :class:`float`, :class:`None`]
        The timer for when the menu times out. Can be :class:`None` for no timeout (defaults to `60.0`)
    
    wrap_in_codeblock: :class:`str`
        The discord codeblock language identifier to wrap your data in (:attr:`ViewMenu.TypeEmbedDynamic` only/defaults to :class:`None`). Example: `ViewMenu(ctx, ..., wrap_in_codeblock='py')`
    """
    
    _active_sessions: List[ViewMenu] = []
    
    def __init__(self, method: Union[Context, discord.Interaction], /, *, menu_type: MenuType, **kwargs):
        super().__init__(method, menu_type, **kwargs)

        self.__buttons: List[ViewButton] = []
        self._stop_initiated = False

        # kwargs
        self.disable_items_on_timeout: bool = kwargs.get('disable_items_on_timeout', True)
        self.remove_items_on_timeout: bool = kwargs.get('remove_items_on_timeout', False)
        self.__timeout: Union[int, float, None] = kwargs.get('timeout', 60.0) # property get/set

        # view
        self.__view = discord.ui.View(timeout=self.__timeout)
        self.__view.on_timeout = self._on_dpy_view_timeout
        self.__view.on_error = self._on_dpy_view_error

        # select
        self.__selects: List[ViewSelect] = []
        self._options_relay_info: Optional[_SelectOptionRelayPayload] = None
        self._gotos: List[ViewSelect.GoTo] = []
    
    def __repr__(self):
        return f'<ViewMenu name={self.name!r} owner={str(self._extract_proper_user(self._method))!r} is_running={self._is_running} timeout={self.timeout} menu_type={self._menu_type.name}>'

    async def _on_dpy_view_timeout(self) -> None:
        self._menu_timed_out = True
        await self.stop(delete_menu_message=self.delete_on_timeout, disable_items=self.disable_items_on_timeout, remove_items=self.remove_items_on_timeout) 
    
    async def _on_dpy_view_error(self, interaction: discord.Interaction, error: Exception, item: discord.ui.Item) -> NoReturn:
        try:
            raise error
        finally:
            await self.stop()
    
    def _get_new_view(self) -> discord.ui.View:
        """Returns a new :class:`discord.ui.View` object with the `timeout` parameter already set along with `on_timeout` and `on_error`"""
        new_view = discord.ui.View(timeout=self.timeout)
        new_view.on_timeout = self._on_dpy_view_timeout
        new_view.on_error = self._on_dpy_view_error
        return new_view
    
    @property
    def selects(self) -> List[ViewSelect]:
        """
        Returns
        -------
        List[:class:`ViewSelect`]: All selects that have been added to the menu
        
            .. added:: v3.1.0
        """
        return self.__selects
    
    @property
    def go_to_selects(self) -> List[ViewSelect.GoTo]:
        """
        Returns
        -------
        List[:class:`ViewSelect.GoTo`]: All go to selects that have been added to the menu
        
            .. added:: v3.1.0
        """
        return self._gotos
    
    @property
    def timeout(self):
        return self.__timeout
    
    @timeout.setter
    def timeout(self, value) -> Union[int, float, None]:
        """A property getter/setter for kwarg `timeout`"""
        if isinstance(value, (int, float, type(None))):
            self.__view.timeout = value
            self.__timeout = value
        else:
            raise IncorrectType(f'"timeout" expected int, float, or None, got {value.__class__.__name__}')
        
    @property
    def buttons(self) -> List[ViewButton]:
        """
        Returns
        -------
        List[:class:`ViewButton`]: All buttons that have been added to the menu
        """
        return self.__buttons
    
    @property
    def buttons_most_clicked(self) -> List[ViewButton]:
        """
        Returns
        -------
        List[:class:`ViewButton`]: The list of buttons on the menu ordered from highest (button with the most clicks) to lowest (button with the least clicks). Can be an empty list if there are no buttons registered to the menu
        """
        return self._sort_buttons(self.__buttons)
    
    @classmethod
    async def quick_start(cls, method: Union[Context, discord.Interaction], /, pages: Sequence[Union[discord.Embed, str]], buttons: Optional[Sequence[ViewButton]]=DEFAULT_BUTTONS) -> ViewMenu:
        """|coro class method|
        
        Start a menu with default settings either with a `menu_type` of `ViewMenu.TypeEmbed` (all values in `pages` are of type `discord.Embed`) or `ViewMenu.TypeText` (all values in `pages` are of type `str`)

        Parameters
        ----------
        method: Union[:class:`discord.ext.commands.Context`, :class:`discord.Interaction`]
            The Context object. You can get this using a command or if you're in a `discord.on_message` event. Also accepts interactions, typically received when using slash commands

        pages: Sequence[Union[:class:`discord.Embed`, :class:`str`]]
            The pages to add

        buttons: Optional[Sequence[:class:`ViewButton`]]
            The buttons to add. If left as `DEFAULT_BUTTONS`, that is equivalent to `ViewButton.all()`
        
        Returns
        -------
        :class:`ViewMenu`: The menu that was started
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `NoPages`: The menu was started when no pages have been added
        - `NoButtons`: Attempted to start the menu when no Buttons have been registered
        - `IncorrectType`: All items in :param:`pages` were not of type :class:`discord.Embed` or :class:`str`
        
            .. added v3.1.0
        """
        menu = cls(method, menu_type=cls._quick_check(pages))
        menu.add_pages(pages) # type: ignore
        menu.add_buttons(ViewButton.all() if buttons is DEFAULT_BUTTONS else buttons) # type: ignore
        await menu.start()
        return menu
    
    def _should_persist(self, button: ViewButton) -> bool:
        """Determine if a link button should stay enabled/remain on the menu when it times out or is stopped

            .. added:: v3.1.0
        """
        return True if all([
            button.custom_id is None,
            button.url,
            button.persist,
            self._stop_initiated
        ]) else False
    
    def _check(self, inter: discord.Interaction) -> bool:
        """Base menu button interaction check. Verifies who (user, everyone, or role) can interact with the button"""
        author_pass = False
        if self._extract_proper_user(inter).id == self.owner.id: author_pass = True
        if self.only_roles: self.all_can_click = False

        if self.only_roles:
            for role in self.only_roles:
                if role in inter.user.roles: # type: ignore / will be :class:`discord.Member`. :attr:`only_roles` will always be `None` because of overridden DM settings so this line will never be reached
                    author_pass = True
                    break
        if self.all_can_click:
            author_pass = True

        return author_pass
    
    async def _handle_event(self, button: ViewButton) -> None:
        """|coro| If an event is set, disable/remove the buttons from the menu when the click requirement has been met"""
        if button.event:
            event_type = button.event.event_type
            event_value = button.event.value
            if button.total_clicks == event_value:
                if event_type == ViewButton.Event._DISABLE:
                    self.disable_button(button)
                
                elif event_type == ViewButton.Event._REMOVE:
                    self.remove_button(button)
                
                await self.refresh_menu_items()
    
    def _remove_director(self, page: Page) -> Page:
        """Removes the page director contents from the page. This is used for :meth:`ViewMenu.update()`"""
        style = self.style
        if style is None:
            style = _DEFAULT_STYLE
        
        escaped_style = re.escape(style)
        STYLE_PATTERN = escaped_style.replace(r'\$', r'\d{1,}').replace(r'\&', r'\d{1,}')
        STYLE_STR_PATTERN = escaped_style.replace(r'\$', r'\d{1,}').replace(r'\&', r'(\d{1,}.*)')
        
        if self.show_page_director:
            # TypeEmbed
            if isinstance(page.embed, discord.Embed) and self._menu_type == ViewMenu.TypeEmbed:
                embed = page.embed
                if embed.footer.text:
                    DIRECTOR_PATTERN = STYLE_PATTERN + r':? '
                    if re.search(DIRECTOR_PATTERN, embed.footer.text):
                        embed.set_footer(text=re.sub(DIRECTOR_PATTERN, '', embed.footer.text), icon_url=embed.footer.icon_url)
                
                return page
            
            # TypeText
            elif isinstance(page.content, str) and self._menu_type == ViewMenu.TypeText:
                if re.search(STYLE_STR_PATTERN, page.content):
                    page.content = re.sub(STYLE_STR_PATTERN, '', page.content).rstrip('\n')
                
                return page

            else:
                raise TypeError(f'_remove_director parameter "page" expected discord.Embed or str, got {page.__class__.__name__}')
        else:
            return page
    
    def set_select_option_relay(self, func: Callable[[NamedTuple], None], *, only: Optional[Sequence[str]]=None) -> None:
        """Set a function to be called with a given set of information when a select option is pressed on the menu. The information passed is `SelectOptionRelayPayload`, a named tuple.
        The named tuple contains the following attributes:

        - `member`: The :class:`discord.Member` object of the person who pressed the option. Could be :class:`discord.User` if the menu was started in a DM
        - `option`: The :class:`discord.SelectOption` that was pressed
        - `menu`: An instance of :class:`ViewMenu` that the select option is operating under

        Parameters
        ----------
        func: Callable[[:class:`NamedTuple`], :class:`None`]
            The function should only contain a single positional argument. Command functions (`@bot.command()`) not supported

        only: Optional[Sequence[:class:`str`]]
            A sequence of :class:`discord.SelectOption` labels associated with the current menu instance. If this is :class:`None`, all select options on the menu will be relayed.
            If set, only presses from the options matching the given labels specified will be relayed

        Raises
        ------
        - `IncorrectType`: The :param:`func` argument provided was not callable
        
            .. added:: v3.1.0
        """
        
        if callable(func):
            self._options_relay_info = _SelectOptionRelayPayload(func, only)
        else:
            raise IncorrectType('When setting the relay for a select option, argument "func" must be callable')
    
    def remove_select_option_relay(self) -> None:
        """Remove the select option relay that's been set
        
            .. added:: v3.1.0
        """
        self._options_relay_info = None
    
    @ensure_not_primed
    def add_select(self, select: ViewSelect) -> None:
        """Add a select category to the menu

        Parameters
        ----------
        select: :class:`ViewSelect`
            The category listing to add
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `ViewMenuException`:  The `menu_type` was not of :attr:`TypeEmbed`. The "embed" parameter in a :class:`Page` was not set. Or both :class:`ViewSelect` and a :class:`ViewSelect.GoTo` were being used

            .. added:: v3.1.0
        """
        if self._gotos:
            raise ViewMenuException('Category selects cannot be used in conjunction with go to selects')
        
        pages = select._view_select_options.values()
        for all_pages in pages:
            for individual_page in all_pages:
                if not individual_page.embed:
                    raise ViewMenuException("The 'embed' parameter in a Page must be set when using selects")
        else:
            if self._menu_type == ViewMenu.TypeEmbed:
                select._menu = self

                # prevent default select options (https://github.com/Defxult/reactionmenu/issues/55)
                for option in select.options:
                    if option.default:
                        option.default = False
                
                self.__view.add_item(select)
                self.__selects.append(select)
            else:
                raise ViewMenuException('Selects can only be added when the menu_type is TypeEmbed')
    
    def remove_select(self, select: ViewSelect) -> None:
        """Remove a select from the menu
        
        Parameters
        ----------
        select: :class:`ViewSelect`
            The select to remove
        
        Raises
        ------
        - `SelectNotFound`: The provided select was not found in the list of selects on the menu
        
            .. added:: v3.1.0
        """
        if select in self.__selects:
            select._menu = None
            self.__view.remove_item(select)
            self.__selects.remove(select)
        else:
            raise SelectNotFound('Cannot remove a select that is not registered')
    
    def remove_all_selects(self) -> None:
        """Remove all selects from the menu
        
            .. added:: v3.1.0
        """
        while self.__selects:
            self.remove_select(self.__selects[0])

    def disable_select(self, select: ViewSelect) -> None:
        """Disable a select on the menu
        
        Parameters
        ----------
        select: :class:`ViewSelect`
            The select to disable
        
        Raises
        ------
        - `SelectNotFound`: The provided select was not found in the list of selects on the menu
        
            .. added:: v3.1.0
        """
        if select in self.__selects:
            select.disabled = True
        else:
            raise SelectNotFound('Cannot disable a select that is not registered')
    
    def disable_all_selects(self) -> None:
        """Disable all selects on the menu
        
            .. added:: v3.1.0
        """
        for select in self.__selects:
            select.disabled = True
    
    def enable_select(self, select: ViewSelect) -> None:
        """Enable the specified select
        
        Parameters
        ----------
        select: :class:`ViewSelect`
            The select to enable
        
        Raises
        ------
        - `SelectNotFound`: The provided select was not found in the list of selects on the menu
        
            .. added:: v3.1.0
        """
        if select in self.__selects:
            select.disabled = False
        else:
            raise SelectNotFound('Cannot enable a select that is not registered')
    
    def enable_all_selects(self) -> None:
        """Enable all selects on the menu
        
            .. added:: v3.1.0
        """
        for select in self.__selects:
            self.enable_select(select)
    
    @overload
    def get_select(self, title: str) -> List[ViewSelect]:
        ...
    
    @overload
    def get_select(self, title: None) -> List[ViewSelect]:
        ...
    
    def get_select(self, title: Union[str, None]) -> List[ViewSelect]:
        """Get a select by it's title (category name) that has been registered to the menu
        
        Parameters
        ----------
        title: Union[:class:`str`, `None`]
            Title of the category
        
        Returns
        -------
        List[:class:`ViewSelect`]: All selects matching the given title
        """
        return [select for select in self.__selects if select.placeholder == title]
    
    @ensure_not_primed
    def add_go_to_select(self, goto: ViewSelect.GoTo) -> None:
        """Add a select where the user can choose which page they'd like to navigate to.

        Parameters
        ----------
        goto: :class:`ViewSelect.GoTo`
            The navigation listing
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `ViewMenuException`:  A :class:`ViewSelect` was already added to the menu. A :class:`ViewSelect` and a :class:`ViewSelect.GoTo` cannot both be used on a single menu 

            .. added:: v3.1.0
        """
        if not self.__selects:
            goto._menu = self
            if goto._page_numbers is ...:
                for n in range(1, len(self._pages) + 1):
                    goto._gt_options.append(discord.SelectOption(label=str(n)))
            self._gotos.append(goto)
            self.__view.add_item(goto)
        else:
            raise ViewMenuException('Category selects cannot be used in conjunction with go to selects')
    
    def enable_go_to_select(self, goto: ViewSelect.GoTo) -> None:
        """Enable the specified go to select

        Parameters
        ----------
        goto: :class:`ViewSelect.GoTo`
            The go to select to enable
        
            .. added:: v3.1.0
        """
        if goto in self._gotos:
            goto.disabled = False
    
    def enable_all_go_to_selects(self) -> None:
        """Enable all go to selects
        
            .. added:: v3.1.0
        """
        for goto in self._gotos:
            goto.disabled = False
    
    def disable_go_to_select(self, goto: ViewSelect.GoTo) -> None:
        """Disable the specified go to select

        Parameters
        ----------
        goto: :class:`ViewSelect.GoTo`
            The go to select to disable
        
            .. added:: v3.1.0
        """
        if goto in self._gotos:
            goto.disabled = True
    
    def disable_all_go_to_selects(self) -> None:
        """Disable all go to selects
        
            .. added:: v3.1.0
        """
        for goto in self._gotos:
            goto.disabled = True
    
    def remove_go_to_select(self, goto: ViewSelect.GoTo) -> None:
        """Remove a go to select from the menu
        
        Parameters
        ----------
        goto: :class:`ViewSelect.GoTo`
            The go to select to remove
        
        Raises
        ------
        - `SelectNotFound`: The provided go to select was not found in the list of selects on the menu
        
            .. added:: v3.1.0
        """
        if goto in self._gotos:
            goto._menu = None
            self.__view.remove_item(goto)
            self._gotos.remove(goto)
        else:
            raise SelectNotFound('Cannot remove a go to select that is not registered')

    def remove_all_go_to_selects(self) -> None:
        """Remove all go to selects from the menu
        
            .. added:: v3.1.0
        """
        while self._gotos:
            self.remove_go_to_select(self._gotos[0])
    
    async def update(self, *, new_pages: Union[List[Union[discord.Embed, str]], None], new_buttons: Union[List[ViewButton], None]) -> None:
        """|coro|
        
        When the menu is running, update the pages or buttons. It should be noted that `ViewSelect`s are not supported here, and they will automatically be removed
        once the menu is updated. This method only supports pages and buttons.
        
        Parameters
        ----------
        new_pages: List[Union[:class:`discord.Embed`, :class:`str`]]
            Pages to *replace* the current pages with. If the menus current `menu_type` is :attr:`ViewMenu.TypeEmbed`, only :class:`discord.Embed` can be used. If :attr:`ViewMenu.TypeText`, only :class:`str` can be used. If you
            don't want to replace any pages, set this to :class:`None`
        
        new_buttons: List[:class:`ViewButton`]
            Buttons to *replace* the current buttons with. Can be an empty list if you want the updated menu to have no buttons. Can also be set to :class:`None` if you don't want to replace any buttons
        
        Raises
        ------
        - `ViewMenuException`: The :class:`ViewButton` custom_id was not recognized or a :class:`ViewButton` with that ID has already been added
        - `TooManyButtons`: There are already 25 buttons on the menu
        - `IncorrectType`: The values in :param:`new_pages` did not match the :class:`ViewMenu`'s `menu_type`. An attempt to use this method when the `menu_type` is :attr:`ViewMenu.TypeEmbedDynamic` which is not allowed. Or
        all :param:`new_buttons` values were not of type :class:`ViewButton`
        """
        if self._is_running:
            # ----------------------- CHECKS -----------------------

            # Note: button count > 25 check is done in :meth:`ViewMenu.add_button`
            
            if new_pages is None and new_buttons is None:
                return

            if self._menu_type not in (ViewMenu.TypeEmbed, ViewMenu.TypeText):
                raise IncorrectType('Updating a menu is only valid for a menu with menu_type ViewMenu.TypeEmbed or ViewMenu.TypeText')
            
            if self._menu_type == ViewMenu.TypeEmbed and new_pages:
                if not all([isinstance(page, discord.Embed) for page in new_pages]):
                    raise IncorrectType('When updating the menu, all values must be of type discord.Embed because the current menu_type is ViewMenu.TypeEmbed')
            
            if self._menu_type == ViewMenu.TypeText and new_pages:
                if not all([isinstance(page, str) for page in new_pages]):
                    raise IncorrectType('When updating the menu, all values must be of type str because the current menu_type is ViewMenu.TypeText')
            
            if isinstance(new_pages, list) and len(new_pages) == 0:
                raise ViewMenuException('new_pages cannot be an empty list. Must be None if no new pages should be added')
            
            # ----------------------- END CHECKS -----------------------

            if new_pages is not None:
                # TypeEmbed
                if self._menu_type == ViewMenu.TypeEmbed:
                    for new_embed_page in new_pages:
                        self._remove_director(Page(embed=new_embed_page)) # type: ignore / contains embed if TypeEmbed
                    
                    self._pages = [Page(embed=e) for e in new_pages.copy()] # type: ignore
                    self._pc = _PageController(self._pages)
                    self._refresh_page_director_info(ViewMenu.TypeEmbed, self._pages)
                
                # TypeText
                else:
                    removed_director_info: List[Page] = []
                    for new_str_page in new_pages.copy():
                        removed_director_info.append(self._remove_director(Page(content=new_str_page))) # type: ignore
                    
                    self._pages = removed_director_info.copy()
                    self._pc = _PageController(self._pages)
                    self._refresh_page_director_info(ViewMenu.TypeText, self._pages)
            else:
                # page controller needs to be reset because even though there are no new pages. the page index is still in the location BEFORE the update
                # EXAMPLE: 5 page menu > click Next button  (on page 2) > update menu no new pages > click Next button (on page 3)
                # that makes no sense and resetting the page controller fixes that issue 
                self._pc = _PageController(self._pages)
            
            kwargs_to_pass = {}
            self.remove_all_selects()

            self.__view.stop()
            self.__view = self._get_new_view()

            # re-using current buttons
            if isinstance(new_buttons, type(None)):
                original_buttons = self.__buttons.copy()
                self.remove_all_buttons()
                for orig_button in original_buttons:
                    self._bypass_primed = True
                    self.add_button(orig_button)
            
            # using new buttons
            elif isinstance(new_buttons, list):
                self.remove_all_buttons()
                if len(new_buttons) >= 1: # empty lists mean all buttons will be removed
                    for new_btn in new_buttons:
                        self._bypass_primed = True
                        self.add_button(new_btn)
            
            kwargs_to_pass['view'] = self.__view
            
            if self._menu_type == ViewMenu.TypeEmbed:
                kwargs_to_pass['embed'] = self._pages[0].embed
            else:
                kwargs_to_pass['content'] = self._pages[0].content
            
            await self._msg.edit(**kwargs_to_pass)
    
    def randomize_button_styles(self) -> None:
        """Set all buttons currently registered to the menu to a random :class:`discord.ButtonStyle` excluding link buttons"""
        all_styles = (
            discord.ButtonStyle.blurple,
            discord.ButtonStyle.green,
            discord.ButtonStyle.gray,
            discord.ButtonStyle.red
        )
        for btn in [button for button in self.__buttons if button.style not in (discord.ButtonStyle.link, discord.ButtonStyle.url)]:
            btn.style = random.choice(all_styles)
    
    def set_button_styles(self, style: discord.ButtonStyle) -> None:
        """Set all buttons currently registered to the menu to the specified :class:`discord.ButtonStyle` excluding link buttons
        
        Parameters
        ----------
        style: :class:`discord.ButtonStyle`
            The button style to set
        """
        for btn in [button for button in self.__buttons if button.style not in (discord.ButtonStyle.link, discord.ButtonStyle.url)]:
            btn.style = style
    
    async def refresh_menu_items(self) -> None:
        """|coro|
        
        When the menu is running, update the message to reflect the buttons/selects that were removed, enabled, or disabled
        """
        if self._is_running:
            current_items = self.__view.children.copy()
            self.remove_all_buttons()
            self.remove_all_selects()
            self.remove_all_go_to_selects()
            self.__view.stop()
            self.__view = self._get_new_view()
            for item in current_items:
                self._bypass_primed = True
                if isinstance(item, discord.ui.Select):
                    if item.__class__.__name__ == "ViewSelect":
                        self.add_select(item) # type: ignore / it's subclassed
                    elif item.__class__.__name__ == "GoTo":
                        self.add_go_to_select(item) # type: ignore / it's subclassed
                elif isinstance(item, ViewButton):
                    self.add_button(item)
                elif isinstance(item, ViewSelect.GoTo):
                    self.add_go_to_select(item)
            await self._msg.edit(view=self.__view)
    
    def remove_button(self, button: ViewButton) -> None:
        """Remove a button from the menu
        
        Parameters
        ----------
        button: :class:`ViewButton`
            The button to remove
        
        Raises
        ------
        - `ButtonNotFound`: The provided button was not found in the list of buttons on the menu
        """
        if button in self.__buttons:
            button._menu = None
            self.__buttons.remove(button)
            self.__view.remove_item(button)
        else:
            raise ButtonNotFound('Cannot remove a button that is not registered')
    
    def remove_all_buttons(self) -> None:
        """Remove all buttons from the menu"""
        # Set persists
        persistent_link_buttons: List[ViewButton] = []
        for btn in self.__buttons:
            if self._should_persist(btn):
                persistent_link_buttons.append(btn)
                continue
        else:
            while self.__buttons:
                self.remove_button(self.__buttons[0])
            else:
                for plb in persistent_link_buttons:
                    self._bypass_primed = True
                    self.add_button(plb)
    
    def disable_button(self, button: ViewButton) -> None:
        """Disable a button on the menu
        
        Parameters
        ----------
        button: :class:`ViewButton`
            The button to disable
        
        Raises
        ------
        - `ButtonNotFound`: The provided button was not found in the list of buttons on the menu
        """
        if button in self.__buttons:
            idx = self.__buttons.index(button)
            self.__buttons[idx].disabled = True
        else:
            raise ButtonNotFound('Cannot disable a button that is not registered')
    
    def disable_all_buttons(self) -> None:
        """Disable all buttons on the menu"""
        for btn in self.__buttons:
            if self._should_persist(btn): continue
            btn.disabled = True
    
    def enable_button(self, button: ViewButton) -> None:
        """Enable the specified button
        
        Parameters
        ----------
        button: :class:`ViewButton`
            The button to enable
        
        Raises
        ------
        - `ButtonNotFound`: The provided button was not found in the list of buttons on the menu
        """
        if button in self.__buttons:
            idx = self.__buttons.index(button)
            self.__buttons[idx].disabled = False
        else:
            raise ButtonNotFound('Cannot enable a button that is not registered')
    
    def enable_all_buttons(self) -> None:
        """Enable all buttons on the menu"""
        for btn in self.__buttons:
            btn.disabled = False
    
    def _button_add_check(self, button: ViewButton) -> None:
        """A set of checks to ensure the proper button is being added"""
        # ensure they are using only the ViewButton and not ReactionMenus :class:`ReactionButton`
        if isinstance(button, ViewButton):

            # ensure the button custom_id is a valid one, but skip this check if its a link button because they dont have custom_ids
            if button.style == discord.ButtonStyle.link:
                pass
            else:
                # Note: this needs to be an re search because of buttons with an ID of "[ID]_[unique ID]"
                if not re.search(ViewButton._RE_IDs, button.custom_id): # type: ignore
                    raise ViewMenuException(f'ViewButton custom_id {button.custom_id!r} was not recognized')
            
            # ensure there are no duplicate custom_ids for the base navigation buttons
            # Note: there's no need to have a check for buttons that are not navigation buttons because they have a unique ID and duplicates of those are allowed
            active_button_ids: List[str] = [btn.custom_id for btn in self.__buttons] # type: ignore
            if button.custom_id in active_button_ids:
                if not all([button.custom_id is None, button.style == discord.ButtonStyle.link]):
                    name = ViewButton._get_id_name_from_id(button.custom_id)
                    raise ViewMenuException(f'A ViewButton with custom_id {name!r} has already been added')
            
            # if the menu_type is TypeText, disallow custom embed buttons
            if button.style != discord.ButtonStyle.link and self._menu_type == ViewMenu.TypeText:
                if button.custom_id == ViewButton.ID_CUSTOM_EMBED:
                    if button.followup and button.followup.embed is not None:
                        raise MenuSettingsMismatch('ViewButton with custom_id ViewButton.ID_CUSTOM_EMBED cannot be used when the menu_type is ViewMenu.TypeText')
            
            # if using a skip button, ensure the skip attribute was set
            if button.custom_id == ViewButton.ID_SKIP and button.skip is None:
                raise ViewMenuException('When attempting to add a button custom_id ViewButton.ID_SKIP, the "skip" kwarg was not set')
            
            # ensure there are no more than 25 buttons
            if len(self.__buttons) >= 25:
                raise TooManyButtons('ViewMenu cannot have more than 25 buttons (discord limitation)')
        else:
            raise IncorrectType(f'When adding a button to the ViewMenu, the button type must be ViewButton, got {button.__class__.__name__}')
    
    def _maybe_unique_id(self, button: ViewButton) -> None:
        """Create a unique ID if the `custom_id` for buttons that are allowed to have duplicates
        
            Note ::
                This excludes link buttons because they don't have a `custom_id`
        """
        if button.custom_id in (ViewButton.ID_CALLER, ViewButton.ID_SEND_MESSAGE, ViewButton.ID_CUSTOM_EMBED, ViewButton.ID_SKIP):
            button.custom_id = f'{button.custom_id}_{id(button)}'
    
    @ensure_not_primed
    def add_button(self, button: ViewButton) -> None:
        """Add a button to the menu
        
        Parameters
        ----------
        button: :class:`ViewButton`
            The button to add
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `MenuSettingsMismatch`: The buttons custom_id was set as :attr:`ViewButton.ID_CUSTOM_EMBED` but the `menu_type` is :attr:`ViewMenu.TypeText`
        - `ViewMenuException`: The custom_id for the button was not recognized or a button with that custom_id has already been added
        - `TooManyButtons`: There are already 25 buttons on the menu
        - `IncorrectType`: Parameter :param:`button` was not of type :class:`ViewButton`
        """
        self._button_add_check(button)
        self._maybe_unique_id(button)

        button._menu = self
        self.__view.add_item(button)
        self.__buttons.append(button)
    
    @ensure_not_primed
    def add_buttons(self, buttons: Sequence[ViewButton]) -> None:
        """Add multiple buttons to the menu at once
        
        Parameters
        ----------
        buttons: Sequence[:class:`ViewButton`]
            The buttons to add
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `MenuSettingsMismatch`: One of the buttons `custom_id` was set as :attr:`ViewButton.ID_CUSTOM_EMBED` but the `menu_type` is :attr:`ViewMenu.TypeText`
        - `ViewMenuException`: The `custom_id` for a button was not recognized or a button with that `custom_id` has already been added
        - `TooManyButtons`: There are already 25 buttons on the menu
        - `IncorrectType`: One or more values supplied in parameter :param:`buttons` was not of type :class:`ViewButton`
        """
        for btn in buttons:
            self.add_button(btn)
    
    def get_button(self, identity: str, *, search_by: Literal['label', 'id', 'name']='label') -> List[ViewButton]:
        """Get a button that has been registered to the menu by it's label, custom_id, or name
        
        Parameters
        ----------
        identity: :class:`str`
            The button label, custom_id, or name
        
        search_by: :class:`str`
            How to search for the button. If "label", it's searched by button labels. If "id", it's searched by it's custom_id. 
            If "name", it's searched by button names
        
        Returns
        -------
        List[:class:`ViewButton`]: The button(s) matching the given identity
        
        Raises
        ------
        - `ViewMenuException`: Parameter :param:`search_by` was not "label", "id", or "name"
        """
        identity = str(identity)
        search_by = str(search_by).lower() # type: ignore

        if search_by == 'label':
            matched_labels: List[ViewButton] = [btn for btn in self.__buttons if btn.label and btn.label == identity]
            return matched_labels
        
        elif search_by == 'id':
            matched_ids: List[ViewButton] = [btn for btn in self.__buttons if btn.custom_id and btn.custom_id.startswith(identity)]
            return matched_ids
        
        elif search_by == 'name':
            matched_names: List[ViewButton] = [btn for btn in self.__buttons if btn.name and btn.name == identity]
            return matched_names
        
        else:
            raise ViewMenuException(f'Parameter "search_by" expected "label", "id", or "name", got {search_by!r}')

    async def _paginate(self, button: ViewButton, inter: discord.Interaction) -> None:
        """|coro| When the button is pressed, handle the pagination process"""
        if not self._check(inter):
            await inter.response.defer()
            return
        
        button._update_statistics(inter.user)
        await self._handle_event(button)
        
        if button.custom_id == ViewButton.ID_PREVIOUS_PAGE:
            await inter.response.edit_message(**self._determine_kwargs(self._pc.prev()))
        
        elif button.custom_id == ViewButton.ID_NEXT_PAGE:
            await inter.response.edit_message(**self._determine_kwargs(self._pc.next()))
        
        elif button.custom_id == ViewButton.ID_GO_TO_FIRST_PAGE:
            await inter.response.edit_message(**self._determine_kwargs(self._pc.first_page()))
        
        elif button.custom_id == ViewButton.ID_GO_TO_LAST_PAGE:
            await inter.response.edit_message(**self._determine_kwargs(self._pc.last_page()))
        
        elif button.custom_id == ViewButton.ID_GO_TO_PAGE:
            await inter.response.defer()
            prompt: discord.Message = await self._msg.channel.send(f'{inter.user.display_name}, what page would you like to go to?') # type: ignore / `.channel` is known at this point
            try:
                selection_message: discord.Message = await inter.client.wait_for('message', check=lambda m: all([m.channel.id == self._msg.channel.id, m.author.id == inter.user.id]), timeout=self.timeout) # type: ignore / `.channel` is known at this point
                page = int(selection_message.content)
            except (asyncio.TimeoutError, ValueError):
                return
            else:
                if 1 <= page <= len(self._pages):
                    self._pc.index = page - 1
                    await self._msg.edit(**self._determine_kwargs(self._pc.current_page))
                    if self.delete_interactions:
                        await prompt.delete()
                        await selection_message.delete()
        
        elif button.custom_id == ViewButton.ID_END_SESSION:
            await self.stop(delete_menu_message=True)
        
        else:
            #* NOTE: Link buttons, aka buttons with a `custom_id` of `None` do not send interactions, so there's no need for an if check

            if button.custom_id.startswith(ViewButton.ID_CALLER): # type: ignore
                if button.followup is None or button.followup.details is None:
                    error_msg = 'ViewButton custom_id was set as ViewButton.ID_CALLER but the "followup" kwarg for that ViewButton was not set ' \
                                'or method ViewButton.Followup.set_caller_details(..) was not called to set the caller information'
                    raise ViewMenuException(error_msg)
                
                func = button.followup.details.func
                args = button.followup.details.args
                kwargs = button.followup.details.kwargs

                # reply now because we don't know how long the users function will take to execute
                await inter.response.defer()

                try:
                    if asyncio.iscoroutinefunction(func): await func(*args, **kwargs) # type: ignore / `func` is already confirmed to be a coroutine
                    else: func(*args, **kwargs)
                except Exception as err:
                    call_failed_error_msg = inspect.cleandoc(
                        f"""
                        The button with custom_id ViewButton.ID_CALLER with the label "{button.label}" raised an error during its execution
                        -> {err.__class__.__name__}: {err}
                        """
                    )
                    raise ViewMenuException(call_failed_error_msg)
                else:
                    if button.followup:
                        # if this executes, the user doesn't want to respond with a message, only with the caller function (already called ^)
                        if all((button.followup.content is None, button.followup.embed is None, button.followup.file is None)):
                            pass
                        else:
                            followup_kwargs = button.followup._to_dict()

                            # inter.followup() has no attribute delete_after/details, so manually delete the key/val pairs to avoid :exc:`TypeError`, got an unexpected kwarg
                            del followup_kwargs['delete_after']
                            del followup_kwargs['details']
                            
                            # if there's no file, remove it to avoid an NoneType error
                            if followup_kwargs['file'] is None:
                                del followup_kwargs['file']
                            
                            followup_message: discord.WebhookMessage = await inter.followup.send(**followup_kwargs)
                            
                            if button.followup.delete_after:
                                await followup_message.delete(delay=button.followup.delete_after)
            
            elif button.custom_id.startswith(ViewButton.ID_SEND_MESSAGE): # type: ignore
                if button.followup is None:
                    raise ViewMenuException('ViewButton custom_id was set as ViewButton.ID_SEND_MESSAGE but the "followup" kwarg for that ViewButton was not set')
                
                # there must be at least 1. cannot send an empty message
                if all((button.followup.content is None, button.followup.embed is None, button.followup.file is None)):
                    raise ViewMenuException('When using a ViewButton with a custom_id of ViewButton.ID_SEND_MESSAGE, the followup message cannot be empty')
                
                followup_kwargs = button.followup._to_dict()

                # inter.followup.send() has no kwarg "details"
                del followup_kwargs['details']
                
                # files are ignored
                del followup_kwargs['file']
                
                # inter.followup.send() has no kwarg "delete_after"
                del followup_kwargs['delete_after']

                # defer instead of inter.response.send_message() so `delete_after` and `allowed_mentions` can be used
                # inter.followup.send() is used instead
                await inter.response.defer()

                sent_message: discord.WebhookMessage = await inter.followup.send(**followup_kwargs)
                if button.followup.delete_after:
                    await sent_message.delete(delay=button.followup.delete_after)
            
            elif button.custom_id.startswith(ViewButton.ID_CUSTOM_EMBED): # type: ignore
                if self._menu_type not in (ViewMenu.TypeEmbed, ViewMenu.TypeEmbedDynamic):
                    raise ViewMenuException('Buttons with custom_id ViewButton.ID_CUSTOM_EMBED can only be used when the menu_type is ViewMenu.TypeEmbed or ViewMenu.TypeEmbedDynamic')
                else:
                    if button.followup is None or button.followup.embed is None:
                        raise ViewMenuException('ViewButton custom_id was set as ViewButton.ID_CUSTOM_EMBED but the "followup" kwargs for that ViewButton was not set or the "embed" kwarg for the followup was not set')
                   
                    await inter.response.edit_message(embed=button.followup.embed)            
            
            elif button.custom_id.startswith(ViewButton.ID_SKIP): # type: ignore
                await inter.response.edit_message(**self._determine_kwargs(self._pc.skip(button.skip)))
            
            else:
                # this shouldn't execute because of :meth:`_button_add_check`, but just in case i missed something, raise the appropriate error
                raise ViewMenuException(f'ViewButton custom_id {button.custom_id!r} was not recognized')

        await self._contact_relay(inter.user, button)

    async def stop(self, *, delete_menu_message: bool=False, disable_items: bool=False, remove_items: bool=False) -> None:
        """|coro|
        
        Stops the process of the menu with the option of deleting the menu's message, removing the buttons, or disabling the buttons upon stop
        
        Parameters
        ----------
        delete_menu_message: :class:`bool`
            Delete the message the menu is operating from

        disable_items: :class:`bool`
            Disable the buttons & selects on the menu
        
        remove_items: :class:`bool`
            Remove the buttons & selects from the menu

        Parameter Hierarchy
        -------------------
        Only one option is available when stopping the menu. If you have multiple parameters as `True`, only one will execute
        - `disable_items` > `remove_items`

        Raises
        ------
        - `discord.DiscordException`: Any exception that can be raised when deleting or editing a message
        """
        if self._is_running:
            self._stop_initiated = True
            try:
                if delete_menu_message:
                    await self._msg.delete()
                else:
                    already_disabled = False
                    if disable_items:
                        self.disable_all_buttons()
                        self.disable_all_selects()
                        self.disable_all_go_to_selects()
                        already_disabled = True
                        await self._msg.edit(view=self.__view)
                    
                    if remove_items and not already_disabled:
                        self.remove_all_buttons()
                        self.remove_all_selects()
                        self.remove_all_go_to_selects()
                        await self._msg.edit(view=self.__view)
            
            except discord.DiscordException as dpy_error:
                raise dpy_error
            
            finally:
                self.__view.stop()
                self._is_running = False

                if self in ViewMenu._active_sessions:
                    ViewMenu._active_sessions.remove(self)
                
                self._on_close_event.set()
                await self._handle_on_timeout()
    
    def _override_dm_settings(self) -> None:
        """If a menu session is in a direct message the following settings are disabled/changed because of discord limitations and resource/safety reasons
        
            .. added:: v3.1.0
        """
        if self.in_dms:
            # Can't delete someone else's message in DMs
            if self.delete_interactions:
                self.delete_interactions = False
            
            # There are no roles in DMs
            if self.only_roles:
                self.only_roles = None
            
            # No point in having an *indefinite* menu in DMs
            if self.timeout is None:
                self.timeout = 60.0
    
    def __generate_viewmenu_payload(self) -> dict:
        """Creates the parameters needed for :meth:`discord.Messageable.send()`
        
            .. added:: v3.1.0
        """
        return {
            "content" : self._pages[0].content if self._pages else None,
            "embed" : self._pages[0].embed if self._pages else discord.utils.MISSING,
            "files" : self._pages[0].files if self._pages else discord.utils.MISSING,
            "view" : self.__view,
            "allowed_mentions" : self.allowed_mentions
        }
    
    def __refresh_select_pages(self, menu_type: MenuType) -> None:
        if self.__selects:
            for vm_select in self.__selects:
                for pages in vm_select._view_select_options.values():
                    self._refresh_page_director_info(menu_type, pages)

    @overload
    async def start(self, *, send_to: Optional[str]=None, reply: bool=False) -> None:
        ...
    
    @overload
    async def start(self, *, send_to: Optional[int]=None, reply: bool=False) -> None:
        ...
    
    @overload
    async def start(self, *, send_to: Optional[discord.TextChannel]=None, reply: bool=False) -> None:
        ...
    
    @overload
    async def start(self, *, send_to: Optional[discord.VoiceChannel]=None, reply: bool=False) -> None:
        ...
    
    @overload
    async def start(self, *, send_to: Optional[discord.Thread]=None, reply: bool=False) -> None:
        ...
    
    @ensure_not_primed
    async def start(self, *, send_to: Optional[Union[str, int, discord.TextChannel, discord.VoiceChannel, discord.Thread]]=None, reply: bool=False) -> None:
        """|coro|
        
        Start the menu
        
        Parameters
        ----------
        send_to: Optional[Union[:class:`str`, :class:`int`, :class:`discord.TextChannel`, :class:`discord.VoiceChannel`, :class:`discord.Thread`]]
			The channel/thread you'd like the menu to start in. Use the channel/threads name, ID, or it's object. Please note that if you intend to use a channel/thread object, using
			method :meth:`discord.Client.get_channel()` (or any other related methods), that channel should be in the same list as if you were to use `ctx.guild.text_channels`
			or `ctx.guild.threads`. This only works on a context guild channel basis. That means a menu instance cannot be created in one guild and the menu itself (:param:`send_to`)
			be sent to another. Whichever guild context the menu was instantiated in, the channels/threads of that guild are the only options for :param:`send_to`

            Note: This parameter is not available if your `method` is a :class:`discord.Interaction`, aka a slash command
        
        reply: :class:`bool`
            Enables the menu message to reply to the message that triggered it. Parameter :param:`send_to` must be :class:`None` if this is `True`. This only pertains to a non-interaction based menu.
        
        Raises
        ------
        - `MenuAlreadyRunning`: Attempted to call method after the menu has already started
        - `NoPages`: The menu was started when no pages have been added
        - `NoButtons`: Attempted to start the menu when no Buttons have been registered
        - `ViewMenuException`: The :class:`ViewMenu`'s `menu_type` was not recognized. There were more than one base navigation buttons. Or a :attr:`ViewButton.ID_CUSTOM_EMBED` button was not correctly formatted
        - `DescriptionOversized`: When using a `menu_type` of :attr:`ViewMenu.TypeEmbedDynamic`, the embed description was over discords size limit
        - `IncorrectType`: Parameter :param:`send_to` was not of the expected type
        - `MenuException`: The channel set in :param:`send_to` was not found
        """
        if ViewMenu._sessions_limit_details.set_by_user:
            can_proceed = await self._handle_session_limits()
            if not can_proceed:
                return
        
        self._override_dm_settings()
        
        # checks
        # Note 1: each at least 1 page check is done in it's own if statement to avoid clashing between pages and custom embeds
        # Note 2: at least 1 page check for add_row is done in "(dynamic menu)"
        
        # ensure at least 1 button exists before starting the menu
        if not self.__buttons: raise NoButtons
        if self._menu_type not in ViewMenu._all_menu_types(): raise ViewMenuException('ViewMenu menu_type not recognized')

        reply_kwargs = self._handle_reply_kwargs(send_to, reply)
        menu_payload = self.__generate_viewmenu_payload()
        
        if isinstance(self._method, Context):
            menu_payload.update(reply_kwargs)

        # add page (normal menu)
        if self._menu_type == ViewMenu.TypeEmbed:
            self._refresh_page_director_info(ViewMenu.TypeEmbed, self._pages)

            navigation_btns = [btn for btn in self.__buttons if btn.custom_id in ViewButton._base_nav_buttons()]

            # an re search is required here because buttons with ID_CUSTOM_EMBED dont have a normal ID, the ID is "8_[unique ID]"
            custom_embed_btns = [btn for btn in self.__buttons if btn.style != discord.ButtonStyle.link and re.search(r'8_\d+', btn.custom_id)] # type: ignore / raw string is compatible

            if all([not self._pages, not custom_embed_btns]):
                raise NoPages

            # normal pages, no custom embeds
            if self._pages and not custom_embed_btns:
                await self._handle_send_to(send_to, menu_payload)
            
            # only custom embeds
            elif not self._pages and custom_embed_btns:
                # since there are only custom embeds, there is no need for base navigation buttons, so remove them if any
                for nav_btn in navigation_btns:
                    if nav_btn in self.__buttons:
                        self.__buttons.remove(nav_btn)
                
                # ensure all custom embed buttons have the proper values set
                for custom_btn in custom_embed_btns:
                    if custom_btn.followup is None or custom_btn.followup.embed is None:
                        raise ViewMenuException('ViewButton custom_id was set as ViewButton.ID_CUSTOM_EMBED but the "followup" kwargs for that ViewButton was not set or the "embed" kwarg for the followup was not set')
                
                # since there are only custom embeds, self._pages is still set to :class:`None`, so set the embed in `.send()` to the first custom embed in the list
                menu_payload['embed'] = custom_embed_btns[0].followup.embed # type: ignore
                await self._handle_send_to(send_to, menu_payload)
            
            # normal pages and custom embeds
            else:
                # since there are custom embeds, ensure there is at least one base navigation button so they can switch between the normal pages and custom embeds
                if not navigation_btns:
                    error_msg = inspect.cleandoc(
                        """
                        Since you've added pages and custom embeds, there needs to be at least one base navigation button. Without one, there's no way to go back to the normal pages in the menu if a custom embed button is pressed.
                        The available base navigation buttons are buttons with the custom_id:
                        - ViewButton.ID_PREVIOUS_PAGE
                        - ViewButton.ID_NEXT_PAGE
                        - ViewButton.ID_GO_TO_FIRST_PAGE
                        - ViewButton.ID_GO_TO_LAST_PAGE
                        - ViewButton.ID_GO_TO_PAGE
                        """
                    )
                    raise ViewMenuException(error_msg)
                else:
                    await self._handle_send_to(send_to, menu_payload)
            
            self.__refresh_select_pages(ViewMenu.TypeEmbed)

        # add row (dynamic menu)
        elif self._menu_type == ViewMenu.TypeEmbedDynamic:
            await self._build_dynamic_pages(send_to, view=self.__view, payload=menu_payload)
        
        # add page (text menu)
        else:
            if not self._pages:
                raise NoPages
            
            self._refresh_page_director_info(ViewMenu.TypeText, self._pages)
            menu_payload['content'] = self._pages[0].content # reassign the first page to show to director information
            await self._handle_send_to(send_to, menu_payload)
        
        self._pc = _PageController(self._pages)
        self._is_running = True
        ViewMenu._active_sessions.append(self)
